with open('CKJ_wide.lib') as fi, open('CKJ_lib.lib', 'w') as fo:
    for line in fi:
        tok = line.split()
        if tok and tok[0] == 'P':
            tok[3:-2:2] = [str(int(float(t) / 1.5)) for t in tok[3:-2:2]]

        print(' '.join(tok), file=fo)



"""Simple script for rebuilding .codespell-ignore-lines

Usage:

cat < /dev/null > .codespell-ignore-lines
pre-commit run --all-files codespell >& /tmp/codespell_errors.txt
python3 tools/codespell_ignore_lines_from_errors.py /tmp/codespell_errors.txt > .codespell-ignore-lines

git diff to review changes, then commit, push.
"""

from __future__ import annotations

import sys


def run(args: list[str]) -> None:
    assert len(args) == 1, "codespell_errors.txt"
    cache = {}
    done = set()
    with open(args[0]) as f:
        lines = f.read().splitlines()

    for line in sorted(lines):
        i = line.find(" ==> ")
        if i > 0:
            flds = line[:i].split(":")
            if len(flds) >= 2:
                filename, line_num = flds[:2]
                if filename not in cache:
                    with open(filename) as f:
                        cache[filename] = f.read().splitlines()
                supp = cache[filename][int(line_num) - 1]
                if supp not in done:
                    print(supp)
                    done.add(supp)


if __name__ == "__main__":
    run(args=sys.argv[1:])



import re, sys
from pathlib import Path

COPYRIGHT = """/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright (C) 2022 KiCad Developers, see AUTHORS.txt for contributors.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 * or you may search the http://www.gnu.org website for the version 2 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

"""

class MockGen:
    def __init__(self, path):
        self.members = []
        self.path = path
        with open(self.path) as f:
            self.lines = f.readlines()

    def __parseHeader(self):
        # group 1: retval and function
        # group 2: arguments
        # group 3: other keywords (const, override...)        
        regex = re.compile("virtual\s+([^\(]+)\(([^\)]*)\)(.*)")

        for line in self.lines:
            virtualFcn = regex.search(line)
            if virtualFcn:
                function = {
                    'name':virtualFcn[1],
                    'args':virtualFcn[2],
                    'extra':virtualFcn[3],
                    }
                self.members.append(function)

    def __generateMockDirective(self, virtualMember):
        if "const" in virtualMember['extra']:
            mockDirective = str("MOCK_CONST_METHOD( ")
        else:
            mockDirective = str("MOCK_METHOD( ")

        fcnName = virtualMember['name'].split()[-1]
        arguments = virtualMember['args'].split(',')
        arity = 0 if len(arguments) == 1 and not arguments[0] else len(arguments)
        mockDirective += fcnName + ", " + str(arity) + ", "

        fcnRetval = virtualMember['name'].strip().removesuffix(fcnName).strip()
        mockDirective += fcnRetval + "("

        if arity != 0:
            for i, arg in enumerate(arguments):
                mockDirective += " " if i == 0 else ", "
                mockDirective += arg.strip()
            mockDirective += " "

        return mockDirective + ") );\n"

    def generateMock(self):
        self.__parseHeader()

        with open('qa/mocks/out.txt', 'w') as output:
            output.write(COPYRIGHT)
            for virtualMember in self.members:
                directive = self.__generateMockDirective(virtualMember)
                output.write(directive)

if __name__ == "__main__":
    header = Path(sys.argv[1])
    if header.exists() and len(sys.argv) == 2:
        mockGen = MockGen(header)
        mockGen.generateMock()
    else:
        print("This script generates mocks for Turtle mock framework")
        print(f"Usage: {sys.argv[0]} path/to/header.hpp")



#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2023 Mark Roszko <mark.roszko@gmail.com>
# Copyright (C) 2023 KiCad Developers
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import pytest
import os

def pytest_report_header(config):
    """Print key environment vars."""
    path_val = os.environ.get("PATH", "")
    pythonpath_val = os.environ.get("PYTHONPATH", "")
    kibuildpaths_val = os.environ.get("KICAD_BUILD_PATHS", "")

    message = "Environment:\n"
    message += f"PATH={path_val}\n"
    message += f"PYTHONPATH={pythonpath_val}\n"
    message += f"KICAD_BUILD_PATHS={kibuildpaths_val}\n"

    return message
    
# We need this on Windows when executing from ctest
if os.name == 'nt':
    for p in os.environ[ 'KICAD_BUILD_PATHS' ].split( ':' ):
        if os.path.isdir( p ):
            os.add_dll_directory( p )


#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright The KiCad Developers, see AUTHORS.txt for contributors.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

"""
PADS Importer QA Tests

These tests verify the PADS importer by:
1. Importing PADS .asc files to KiCad format
2. Exporting gerbers from the imported board
3. Comparing exported gerbers against reference gerbers
4. Reporting diff percentages per layer

Environment Variables:
    PADS_QA_SAVE_ALL_DIFFS: Set to "1" to save diff images for all layers,
                            not just failing ones. Useful for debugging.
"""

import utils
from pathlib import Path
import pytest
import json
import os
from conftest import KiTestFixture
from dataclasses import dataclass, field
from typing import Dict, List, Optional


def should_save_all_diffs() -> bool:
    """Check if we should save diff images for all layers, not just failing ones."""
    return os.environ.get( "PADS_QA_SAVE_ALL_DIFFS", "" ) == "1"


# Map KiCad layer names to gerber export layer arguments
KICAD_LAYER_TO_GERBER_ARG = {
    "F.Cu": "F.Cu",
    "B.Cu": "B.Cu",
    "In1.Cu": "In1.Cu",
    "In2.Cu": "In2.Cu",
    "In3.Cu": "In3.Cu",
    "In4.Cu": "In4.Cu",
    "In5.Cu": "In5.Cu",
    "In6.Cu": "In6.Cu",
    "F.Mask": "F.Mask",
    "B.Mask": "B.Mask",
    "F.Paste": "F.Paste",
    "B.Paste": "B.Paste",
    "F.Silkscreen": "F.Silkscreen",
    "B.Silkscreen": "B.Silkscreen",
}

# Map KiCad layer names to gerber export file suffixes
KICAD_LAYER_TO_SUFFIX = {
    "F.Cu": "-F_Cu.gtl",
    "B.Cu": "-B_Cu.gbl",
    "In1.Cu": "-In1_Cu.g2",
    "In2.Cu": "-In2_Cu.g3",
    "In3.Cu": "-In3_Cu.g4",
    "In4.Cu": "-In4_Cu.g5",
    "In5.Cu": "-In5_Cu.g6",
    "In6.Cu": "-In6_Cu.g7",
    "F.Mask": "-F_Mask.gts",
    "B.Mask": "-B_Mask.gbs",
    "F.Paste": "-F_Paste.gtp",
    "B.Paste": "-B_Paste.gbp",
    "F.Silkscreen": "-F_Silkscreen.gto",
    "B.Silkscreen": "-B_Silkscreen.gbo",
}


@dataclass
class LayerResult:
    """Result of comparing a single layer"""
    layer_name: str
    reference_file: str
    generated_file: str
    diff_percentage: float
    passed: bool
    diff_image: Optional[str] = None


@dataclass
class DrillResult:
    """Result of comparing drill files"""
    reference_file: str
    generated_file: str
    diff_percentage: float
    passed: bool
    hole_count_reference: int = 0
    hole_count_generated: int = 0
    diff_image: Optional[str] = None


@dataclass
class TestCaseResult:
    """Result of testing a complete design"""
    name: str
    passed: bool
    layer_results: List[LayerResult] = field( default_factory=list )
    drill_results: List[DrillResult] = field( default_factory=list )
    import_report: Optional[Dict] = None
    error: Optional[str] = None


def get_pads_test_cases() -> List[str]:
    """Get list of available PADS test cases."""
    test_data_dir = Path( __file__ ).parent.parent.parent / "data" / "pcbnew" / "plugins" / "pads"

    if not test_data_dir.exists():
        return []

    test_cases = []

    for item in test_data_dir.iterdir():
        if item.is_dir() and ( item / "config.json" ).exists():
            test_cases.append( item.name )

    return test_cases


def load_test_config( test_case: str ) -> Dict:
    """Load the configuration for a test case."""
    test_data_dir = Path( __file__ ).parent.parent.parent / "data" / "pcbnew" / "plugins" / "pads"
    config_path = test_data_dir / test_case / "config.json"

    with open( config_path, "r" ) as f:
        return json.load( f )


def get_test_data_path( test_case: str ) -> Path:
    """Get the path to test data for a test case."""
    return Path( __file__ ).parent.parent.parent / "data" / "pcbnew" / "plugins" / "pads" / test_case


class PadsImporterHelper:
    """Helper class for PADS importer tests"""

    def __init__( self, kitest: KiTestFixture, test_case: str ):
        self.kitest = kitest
        self.test_case = test_case
        self.config = load_test_config( test_case )
        self.test_data_path = get_test_data_path( test_case )
        self.output_dir = kitest.get_output_path( f"cli/pads_importer/{test_case}/" )

    def import_pads_file( self ) -> tuple:
        """Import the PADS file and return (output_path, report, return_code)"""
        source_file = self.test_data_path / self.config["source_file"]
        output_path = self.output_dir / f"{self.test_case}.kicad_pcb"
        report_path = self.output_dir / "import_report.json"

        if output_path.exists():
            output_path.unlink()

        if report_path.exists():
            report_path.unlink()

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "--format", "pads",
            "--report-format", "json",
            "--report-file", str( report_path ),
            str( source_file ),
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        report = None

        if report_path.exists():
            with open( report_path, "r" ) as f:
                report = json.load( f )

        return output_path, report, return_code

    def export_gerbers( self, board_path: Path ) -> Dict[str, Path]:
        """Export gerbers from the imported board."""
        gerber_output_dir = self.output_dir / "gerbers"
        gerber_output_dir.mkdir( parents=True, exist_ok=True )

        # Build list of all layers to export at once
        layers_to_export = []

        for ref_gerber, kicad_layer in self.config["layer_mapping"].items():
            if kicad_layer in KICAD_LAYER_TO_GERBER_ARG:
                layers_to_export.append( KICAD_LAYER_TO_GERBER_ARG[kicad_layer] )

        if not layers_to_export:
            return {}

        # Use the plural 'gerbers' command which exports all layers at once
        command = [
            utils.kicad_cli(),
            "pcb", "export", "gerbers",
            "--no-x2",
            "--layers", ",".join( layers_to_export ),
            "-o", str( gerber_output_dir ) + "/",
            str( board_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        exported_gerbers = {}

        # Map reference gerbers to exported files
        for ref_gerber, kicad_layer in self.config["layer_mapping"].items():
            if kicad_layer not in KICAD_LAYER_TO_SUFFIX:
                continue

            suffix = KICAD_LAYER_TO_SUFFIX[kicad_layer]
            expected_file = gerber_output_dir / ( board_path.stem + suffix )

            if expected_file.exists():
                exported_gerbers[ref_gerber] = expected_file

        return exported_gerbers

    def compare_layer( self, ref_gerber: str, generated_gerber: Path ) -> LayerResult:
        """Compare a generated gerber against reference."""
        reference_path = self.test_data_path / "reference_gerbers" / ref_gerber

        if not reference_path.exists():
            return LayerResult(
                layer_name=self.config["layer_mapping"][ref_gerber],
                reference_file=str( reference_path ),
                generated_file=str( generated_gerber ),
                diff_percentage=100.0,
                passed=False
            )

        diff_output = self.output_dir / "diffs" / f"{ref_gerber}_diff.png"
        diff_output.parent.mkdir( parents=True, exist_ok=True )

        # First, get JSON output for statistics
        command_json = [
            utils.kicad_cli(),
            "gerber", "diff",
            "--format", "json",
            str( reference_path ),
            str( generated_gerber )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command_json )

        diff_percentage = 100.0

        # Return code 1 means files are different, 0 means identical
        if return_code in ( 0, 1 ):
            json_start = stdout.find( "{" )
            json_end = stdout.rfind( "}" ) + 1

            if json_start >= 0 and json_end > json_start:
                try:
                    diff_result = json.loads( stdout[json_start:json_end] )
                    diff_percentage = diff_result.get( "total_diff_percent", 100.0 )
                except json.JSONDecodeError:
                    pass

        threshold = self.config.get( "pass_threshold", 5.0 )
        passed = diff_percentage <= threshold

        # Generate PNG diff image if there are differences and we need it
        if should_save_all_diffs() or not passed:
            command_png = [
                utils.kicad_cli(),
                "gerber", "diff",
                "--format", "png",
                "-o", str( diff_output ),
                str( reference_path ),
                str( generated_gerber )
            ]

            utils.run_and_capture( command_png )

        return LayerResult(
            layer_name=self.config["layer_mapping"][ref_gerber],
            reference_file=str( reference_path ),
            generated_file=str( generated_gerber ),
            diff_percentage=diff_percentage,
            passed=passed,
            diff_image=str( diff_output ) if diff_output.exists() else None
        )

    def export_drill( self, board_path: Path ) -> Optional[Path]:
        """Export drill files from the imported board."""
        drill_output_dir = self.output_dir / "drill"
        drill_output_dir.mkdir( parents=True, exist_ok=True )

        command = [
            utils.kicad_cli(),
            "pcb", "export", "drill",
            "--format", "excellon",
            "-o", str( drill_output_dir ) + "/",
            str( board_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        # Look for the generated drill file
        expected_file = drill_output_dir / ( board_path.stem + ".drl" )

        if expected_file.exists():
            return expected_file

        # Try PTH specific file
        pth_file = drill_output_dir / ( board_path.stem + "-PTH.drl" )

        if pth_file.exists():
            return pth_file

        return None

    def compare_drill( self, ref_drill: str, generated_drill: Path ) -> DrillResult:
        """Compare a generated drill file against reference."""
        reference_path = self.test_data_path / "reference_gerbers" / ref_drill

        if not reference_path.exists():
            return DrillResult(
                reference_file=str( reference_path ),
                generated_file=str( generated_drill ),
                diff_percentage=100.0,
                passed=False
            )

        diff_output = self.output_dir / "diffs" / f"{ref_drill}_diff.png"
        diff_output.parent.mkdir( parents=True, exist_ok=True )

        # Get JSON output for statistics
        command_json = [
            utils.kicad_cli(),
            "gerber", "diff",
            "--format", "json",
            str( reference_path ),
            str( generated_drill )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command_json )

        diff_percentage = 100.0
        hole_count_ref = 0
        hole_count_gen = 0

        # Return code 1 means files are different, 0 means identical
        if return_code in ( 0, 1 ):
            json_start = stdout.find( "{" )
            json_end = stdout.rfind( "}" ) + 1

            if json_start >= 0 and json_end > json_start:
                try:
                    diff_result = json.loads( stdout[json_start:json_end] )
                    diff_percentage = diff_result.get( "total_diff_percent", 100.0 )
                except json.JSONDecodeError:
                    pass

        threshold = self.config.get( "pass_threshold", 5.0 )
        passed = diff_percentage <= threshold

        # Generate PNG diff image if there are differences and we need it
        if should_save_all_diffs() or not passed:
            command_png = [
                utils.kicad_cli(),
                "gerber", "diff",
                "--format", "png",
                "-o", str( diff_output ),
                str( reference_path ),
                str( generated_drill )
            ]

            utils.run_and_capture( command_png )

        return DrillResult(
            reference_file=str( reference_path ),
            generated_file=str( generated_drill ),
            diff_percentage=diff_percentage,
            passed=passed,
            hole_count_reference=hole_count_ref,
            hole_count_generated=hole_count_gen,
            diff_image=str( diff_output ) if diff_output.exists() else None
        )

    def run_test( self ) -> TestCaseResult:
        """Run the complete import and comparison test."""
        result = TestCaseResult( name=self.test_case, passed=False )

        board_path, import_report, return_code = self.import_pads_file()
        result.import_report = import_report

        if return_code != 0 or not board_path.exists():
            result.error = f"Import failed with return code {return_code}"
            return result

        exported_gerbers = self.export_gerbers( board_path )

        if not exported_gerbers:
            result.error = "No gerbers were exported"
            return result

        all_passed = True

        for ref_gerber, generated_gerber in exported_gerbers.items():
            layer_result = self.compare_layer( ref_gerber, generated_gerber )
            result.layer_results.append( layer_result )

            if not layer_result.passed:
                all_passed = False

        # Compare drill files if configured
        drill_files = self.config.get( "drill_files", [] )

        if drill_files:
            generated_drill = self.export_drill( board_path )

            if generated_drill:
                for ref_drill in drill_files:
                    drill_result = self.compare_drill( ref_drill, generated_drill )
                    result.drill_results.append( drill_result )

                    if not drill_result.passed:
                        all_passed = False

        result.passed = all_passed
        return result

    def save_results( self, result: TestCaseResult ) -> Path:
        """Save test results to JSON file."""
        results_path = self.output_dir / "results.json"

        results_dict = {
            "name": result.name,
            "passed": result.passed,
            "error": result.error,
            "import_report": result.import_report,
            "layer_results": [
                {
                    "layer_name": lr.layer_name,
                    "reference_file": lr.reference_file,
                    "generated_file": lr.generated_file,
                    "diff_percentage": lr.diff_percentage,
                    "passed": lr.passed,
                    "diff_image": lr.diff_image
                }
                for lr in result.layer_results
            ],
            "drill_results": [
                {
                    "reference_file": dr.reference_file,
                    "generated_file": dr.generated_file,
                    "diff_percentage": dr.diff_percentage,
                    "passed": dr.passed,
                    "hole_count_reference": dr.hole_count_reference,
                    "hole_count_generated": dr.hole_count_generated,
                    "diff_image": dr.diff_image
                }
                for dr in result.drill_results
            ]
        }

        with open( results_path, "w" ) as f:
            json.dump( results_dict, f, indent=2 )

        return results_path


# Get test cases at module load time
PADS_TEST_CASES = get_pads_test_cases()


@pytest.mark.skipif( len( PADS_TEST_CASES ) == 0,
                     reason="No PADS test data available" )
class TestPadsImporter:
    """Parametrized tests for PADS importer"""

    @pytest.mark.parametrize( "test_case", PADS_TEST_CASES )
    def test_pads_import_and_compare( self, kitest: KiTestFixture, test_case: str ):
        """Test importing a PADS file and comparing gerbers and drill files."""
        helper = PadsImporterHelper( kitest, test_case )
        result = helper.run_test()
        results_path = helper.save_results( result )

        if result.error:
            pytest.fail( f"Test failed: {result.error}" )

        # Attach diff images for failed layers
        for layer_result in result.layer_results:
            if not layer_result.passed:
                if layer_result.diff_image:
                    kitest.add_attachment( layer_result.diff_image )

        # Attach diff images for failed drill comparisons
        for drill_result in result.drill_results:
            if not drill_result.passed:
                if drill_result.diff_image:
                    kitest.add_attachment( drill_result.diff_image )

        failed_layers = [lr for lr in result.layer_results if not lr.passed]
        failed_drills = [dr for dr in result.drill_results if not dr.passed]

        if failed_layers or failed_drills:
            msg = f"Comparison failures for {test_case}:\n"

            if failed_layers:
                msg += "  Layer failures:\n"

                for lr in failed_layers:
                    msg += f"    {lr.layer_name}: {lr.diff_percentage:.2f}% diff (threshold: {helper.config.get('pass_threshold', 5.0)}%)\n"

            if failed_drills:
                msg += "  Drill failures:\n"

                for dr in failed_drills:
                    ref_name = Path( dr.reference_file ).name
                    msg += f"    {ref_name}: {dr.diff_percentage:.2f}% diff (threshold: {helper.config.get('pass_threshold', 5.0)}%)\n"

            pytest.fail( msg )

    @pytest.mark.parametrize( "test_case", PADS_TEST_CASES )
    def test_pads_import_only( self, kitest: KiTestFixture, test_case: str ):
        """Test that PADS import completes without error."""
        helper = PadsImporterHelper( kitest, test_case )
        board_path, import_report, return_code = helper.import_pads_file()

        assert return_code == 0, f"Import failed with return code {return_code}"
        assert board_path.exists(), f"Output file not created: {board_path}"
        assert import_report is not None, "Import report not generated"
        assert "statistics" in import_report, "Import report missing statistics"





#
# Example python script to generate a BOM from a KiCad generic netlist
#
# Example: Sorted and Grouped CSV BOM
#

"""
    @package
    Output: CSV (comma-separated)
    Grouped By: Value, Footprint, DNP, specified extra fields
    Sorted By: Reference
    Fields: #, Reference, Qty, Value, Footprint, DNP, specified extra fields

    Outputs components grouped by Value, Footprint, and specified extra fields.
    Extra fields can be passed as command line arguments at the end, one field per argument.

    Command line:
    python "pathToFile/bom_csv_grouped_extra.py" "%I" "%O.csv" "Extra_Field1" "Extra_Field2"
"""

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import kicad_utils
import csv
import sys

# Get extra fields from the command line
extra_fields = sys.argv[3:]

comp_fields = ['Value', 'Footprint', 'DNP'] + extra_fields
header_names = ['#', 'Reference', 'Qty'] + comp_fields

def getComponentString(comp, field_name):
    if field_name == "Value":
        return comp.getValue()
    elif field_name == "Footprint":
        return comp.getFootprint()
    elif field_name == "DNP":
        return comp.getDNPString()
    elif field_name == "Datasheet":
        return comp.getDatasheet()
    else:
        return comp.getField( field_name )

def myEqu(self, other):
    """myEqu is a more advanced equivalence function for components which is
    used by component grouping. Normal operation is to group components based
    on their Value and Footprint.

    In this example of a more advanced equivalency operator we also compare the
    Footprint, Value, DNP and all extra fields passed from the command line. If
    these fields are not used in some parts they will simply be ignored (they
    will match as both will be empty strings).

    """
    result = True
    for field_name in comp_fields:
        if getComponentString(self, field_name) != getComponentString(other, field_name):
            result = False

    return result

# Override the component equivalence operator - it is important to do this
# before loading the netlist, otherwise all components will have the original
# equivalency operator.
kicad_netlist_reader.comp.__eq__ = myEqu

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = kicad_utils.open_file_writeUTF8(sys.argv[2], 'w')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

# Create a new csv writer object to use as the output formatter
out = csv.writer(f, lineterminator='\n', delimiter=',', quotechar='\"', quoting=csv.QUOTE_ALL)

# Output a CSV header
out.writerow(header_names)

components = net.getInterestingComponents( excludeBOM=True )

# Get all of the components in groups of matching parts + values
# (see kicad_netlist_reader.py)
grouped = net.groupComponents(components)

# Output all of the component information
index = 1
for group in grouped:
    refs = ""
    refs_l = []

    # Add the reference of every component in the group and keep a reference
    # to the component so that the other data can be filled in once per group
    for component in group:
        refs_l.append( component.getRef() )
        c = component

    refs = ", ".join(refs_l)

    # Fill in the component groups common data
    row = []
    row.append( index )
    row.append( refs )
    row.append( len(group) )

    # Add the values of component-specific data
    for field_name in comp_fields:
        row.append( getComponentString( c, field_name ) )

    out.writerow(row)

    index += 1

if not canOpenFile:
    sys.exit(1)



"""
Tests for the KiCad WebView test server.

Run with:  pytest utils/webview_test/server/tests/
"""

from __future__ import annotations

import base64
import json
import socketserver
import sys
import threading
import urllib.error
import urllib.request
from pathlib import Path
from typing import Generator

import pytest
import zstandard

# Allow importing the server module from its parent directory.
sys.path.insert(0, str(Path(__file__).parent.parent))

from webview_test_server import (
    _ServerState,
    _build_handler,
    _content_type_for_command,
    _encode_bytes,
    _extract_asset_name,
)

_DATA_DIR = Path(__file__).parent.parent / "data"
_CONFIG_FILE = _DATA_DIR / "parts.json"


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture(scope="module")
def server_state() -> _ServerState:
    return _ServerState(config_path=_CONFIG_FILE)


class _ReuseAddrServer(socketserver.TCPServer):
    allow_reuse_address = True


@pytest.fixture(scope="module")
def live_server() -> Generator[int, None, None]:
    """Start a live HTTP server on an OS-assigned port and yield that port."""
    state = _ServerState(config_path=_CONFIG_FILE)
    handler_cls = _build_handler(state)

    with _ReuseAddrServer(("127.0.0.1", 0), handler_cls) as httpd:
        port = httpd.server_address[1]
        thread = threading.Thread(target=httpd.serve_forever, daemon=True)
        thread.start()
        yield port
        httpd.shutdown()


# ---------------------------------------------------------------------------
# Provider metadata
# ---------------------------------------------------------------------------


_BASE_URL = "http://localhost:8080"


class TestProviderMetadata:
    def test_returns_bytes(self, server_state: _ServerState) -> None:
        result = server_state.build_provider_metadata(_BASE_URL)
        assert isinstance(result, bytes)
        assert len(result) > 0

    def test_valid_json(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))
        assert isinstance(metadata, dict)

    def test_required_fields_present(self, server_state: _ServerState) -> None:
        required = [
            "provider_name",
            "provider_version",
            "api_base_url",
            "panel_url",
            "auth",
            "capabilities",
            "max_download_bytes",
            "supported_asset_types",
        ]
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))

        for field in required:
            assert field in metadata, f"Required field '{field}' missing from provider metadata"

    def test_capabilities(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))
        caps = metadata["capabilities"]
        assert caps.get("web_ui_v1") is True, "web_ui_v1 must be true"
        assert caps.get("inline_payloads_v1") is True, "inline_payloads_v1 must be true"

    def test_auth_type_none(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))
        assert metadata["auth"]["type"] == "none"

    def test_allow_insecure_localhost(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))
        assert metadata.get("allow_insecure_localhost") is True

    def test_supported_asset_types(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))
        types = metadata["supported_asset_types"]
        assert len(types) >= 1
        for expected in ("symbol", "footprint"):
            assert expected in types

    def test_max_download_bytes_positive(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata(_BASE_URL))
        assert metadata["max_download_bytes"] > 0

    def test_base_url_reflected_in_urls(self, server_state: _ServerState) -> None:
        metadata = json.loads(server_state.build_provider_metadata("http://localhost:12345"))
        assert "12345" in metadata["api_base_url"]
        assert "12345" in metadata["panel_url"]


# ---------------------------------------------------------------------------
# Asset encoding
# ---------------------------------------------------------------------------


class TestEncodeBytes:
    def test_roundtrip_text(self) -> None:
        original = b"Hello, KiCad!"
        encoded = _encode_bytes(original)
        compressed = base64.b64decode(encoded)
        decompressor = zstandard.ZstdDecompressor()
        assert decompressor.decompress(compressed) == original

    def test_roundtrip_binary(self) -> None:
        original = bytes(range(256))
        encoded = _encode_bytes(original)
        compressed = base64.b64decode(encoded)
        decompressor = zstandard.ZstdDecompressor()
        assert decompressor.decompress(compressed) == original

    def test_produces_valid_base64(self) -> None:
        encoded = _encode_bytes(b"test payload")
        base64.b64decode(encoded)  # must not raise

    def test_empty_input(self) -> None:
        encoded = _encode_bytes(b"")
        compressed = base64.b64decode(encoded)
        decompressor = zstandard.ZstdDecompressor()
        assert decompressor.decompress(compressed) == b""

    def test_compression_reduces_repetitive_data(self) -> None:
        original = b"AAAA" * 1000
        encoded = _encode_bytes(original)
        assert len(base64.b64decode(encoded)) < len(original)


# ---------------------------------------------------------------------------
# Asset name extraction
# ---------------------------------------------------------------------------


class TestExtractAssetName:
    def test_kicad_symbol(self, tmp_path: Path) -> None:
        sym = tmp_path / "r.kicad_sym"
        sym.write_text('(kicad_symbol_lib\n  (symbol "R"\n  )\n)')
        assert _extract_asset_name(sym, "DL_SYMBOL") == "R"

    def test_kicad_symbol_quoted_name_with_spaces(self, tmp_path: Path) -> None:
        sym = tmp_path / "cap.kicad_sym"
        sym.write_text('(kicad_symbol_lib\n  (symbol "C_Small"\n  )\n)')
        assert _extract_asset_name(sym, "DL_SYMBOL") == "C_Small"

    def test_kicad_footprint(self, tmp_path: Path) -> None:
        fp = tmp_path / "r.kicad_mod"
        fp.write_text('(footprint "R_0603_1608Metric" (layer "F.Cu"))')
        assert _extract_asset_name(fp, "DL_FOOTPRINT") == "R_0603_1608Metric"

    def test_symbol_fallback_to_stem(self, tmp_path: Path) -> None:
        sym = tmp_path / "my_symbol.kicad_sym"
        sym.write_text("(kicad_symbol_lib)")
        assert _extract_asset_name(sym, "DL_SYMBOL") == "my_symbol"

    def test_binary_file_fallback(self, tmp_path: Path) -> None:
        step = tmp_path / "R_0603.step"
        step.write_bytes(b"\x00\x01\x02\x03ISO-10303-21")
        assert _extract_asset_name(step, "DL_3DMODEL") == "R_0603"

    def test_spice_falls_back_to_stem(self, tmp_path: Path) -> None:
        cir = tmp_path / "resistor.cir"
        cir.write_text("* SPICE model\n.SUBCKT R A B\n.ENDS\n")
        assert _extract_asset_name(cir, "DL_SPICE") == "resistor"


# ---------------------------------------------------------------------------
# Content type mapping
# ---------------------------------------------------------------------------


class TestContentTypeForCommand:
    def test_dl_symbol(self) -> None:
        assert _content_type_for_command("DL_SYMBOL") == "KICAD_SYMBOL_V1"

    def test_dl_footprint(self) -> None:
        assert _content_type_for_command("DL_FOOTPRINT") == "KICAD_FOOTPRINT_V1"

    def test_dl_spice(self) -> None:
        assert _content_type_for_command("DL_SPICE") == "KICAD_SPICE_MODEL_V1"

    def test_dl_3dmodel(self) -> None:
        assert _content_type_for_command("DL_3DMODEL") == "KICAD_3D_MODEL_STEP"

    def test_case_insensitive(self) -> None:
        assert _content_type_for_command("dl_symbol") == "KICAD_SYMBOL_V1"

    def test_unknown_command(self) -> None:
        assert _content_type_for_command("UNKNOWN_CMD") == "UNKNOWN"


# ---------------------------------------------------------------------------
# Parts loading
# ---------------------------------------------------------------------------


class TestPartsLoading:
    def test_loads_expected_parts(self, server_state: _ServerState) -> None:
        page_html = server_state.page.decode("utf-8")
        assert "Resistor 0603 1k" in page_html
        assert "Capacitor 0603 100nF" in page_html

    def test_page_contains_session_logic(self, server_state: _ServerState) -> None:
        page_html = server_state.page.decode("utf-8")
        assert "NEW_SESSION" in page_html
        assert "kiclient" in page_html

    def test_page_contains_dl_component(self, server_state: _ServerState) -> None:
        page_html = server_state.page.decode("utf-8")
        assert "DL_COMPONENT" in page_html

    def test_page_is_valid_html(self, server_state: _ServerState) -> None:
        page_html = server_state.page.decode("utf-8")
        assert page_html.startswith("<!DOCTYPE html>")
        assert "</html>" in page_html

    def test_missing_config_raises(self, tmp_path: Path) -> None:
        with pytest.raises(FileNotFoundError):
            _ServerState(config_path=tmp_path / "nonexistent.json")

    def test_page_uses_current_remote_login_protocol(self, server_state: _ServerState) -> None:
        page_html = server_state.page.decode("utf-8")
        assert 'sendRpcCommand("REMOTE_LOGIN", { interactive: true })' in page_html
        assert "const LOGIN_ENDPOINT" not in page_html
        assert "mockAuthentication(" not in page_html

    def test_page_registers_waiter_for_new_session_ack(self, server_state: _ServerState) -> None:
        page_html = server_state.page.decode("utf-8")
        assert "const msgId = ++messageCounter;" in page_html
        assert "waitForResponse(msgId, 4000).catch(() => {});" in page_html


# ---------------------------------------------------------------------------
# HTTP endpoint integration tests
# ---------------------------------------------------------------------------


class TestHttpEndpoints:
    def test_root_returns_200(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/"
        with urllib.request.urlopen(url) as resp:
            assert resp.status == 200

    def test_root_content_type(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/"
        with urllib.request.urlopen(url) as resp:
            assert "text/html" in resp.headers.get("Content-Type", "")

    def test_root_contains_parts(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/"
        with urllib.request.urlopen(url) as resp:
            body = resp.read().decode("utf-8")
        assert "Resistor 0603 1k" in body

    def test_provider_metadata_endpoint(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/.well-known/kicad-remote-provider"
        with urllib.request.urlopen(url) as resp:
            assert resp.status == 200
            content_type = resp.headers.get("Content-Type", "")
            assert "application/json" in content_type
            metadata = json.loads(resp.read())
        assert metadata["capabilities"]["web_ui_v1"] is True
        assert metadata["auth"]["type"] == "none"

    def test_provider_metadata_port_matches_server(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/.well-known/kicad-remote-provider"
        with urllib.request.urlopen(url) as resp:
            metadata = json.loads(resp.read())
        assert str(live_server) in metadata["api_base_url"]
        assert str(live_server) in metadata["panel_url"]

    def test_unknown_path_returns_404(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/no-such-path"
        with pytest.raises(urllib.error.HTTPError) as exc_info:
            urllib.request.urlopen(url)
        assert exc_info.value.code == 404

    def test_index_html_alias(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/index.html"
        with urllib.request.urlopen(url) as resp:
            assert resp.status == 200

    def test_login_page(self, live_server: int) -> None:
        url = f"http://127.0.0.1:{live_server}/login"
        with urllib.request.urlopen(url) as resp:
            assert resp.status == 200
            assert b"Sign In" in resp.read()



# This program source code file is part of KiCad, a free EDA CAD application.
# Copyright The KiCad Developers, see AUTHORS.TXT for contributors.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, you may find one here:
# https://www.gnu.org/licenses/gpl-3.0.en.html

"""
KiCad HTTP / Remote Provider Test Server
========================================

Local fixture server for two paths:

- legacy HTTP symbol library endpoints under `/v1/`
- new remote provider metadata/manifest/download endpoints

The provider flow includes a provider-owned embedded page plus lightweight
OAuth and download stubs so KiCad can exercise the full panel flow without a
third-party service.
"""

from __future__ import annotations

import argparse
import base64
import hashlib
import json
from typing import Dict, List, Optional

import uvicorn
from fastapi import FastAPI, Form, HTTPException, Request
from fastapi.responses import PlainTextResponse, RedirectResponse, Response
from pydantic import BaseModel

app = FastAPI()


def base_url(request: Request) -> str:
    return str(request.base_url).rstrip("/")


class Category(BaseModel):
    id: str
    name: str
    description: Optional[str] = None


class PartsChooser(BaseModel):
    id: str
    name: str
    description: Optional[str] = None


class PartField(BaseModel):
    value: str
    visible: Optional[str] = "true"


class DetailedPart(BaseModel):
    id: str
    name: Optional[str] = None
    symbolIdStr: str
    exclude_from_bom: str
    exclude_from_board: str
    exclude_from_sim: str
    fields: Dict[str, PartField]


categories = [
    Category(id="1", name="Resistors", description="Resistors"),
]

parts_by_category: Dict[str, List[PartsChooser]] = {
    "1": [
        PartsChooser(
            id="1",
            name="RC0603FR-0710KL",
            description="10 kOhms ±1% 0.1W, 1/10W Chip Resistor 0603 (1608 Metric) Moisture Resistant Thick Film",
        )
    ]
}

detailed_parts = {
    "1": DetailedPart(
        id="1",
        name="RC0603FR-0710KL",
        symbolIdStr="Device:R",
        exclude_from_bom="False",
        exclude_from_board="False",
        exclude_from_sim="True",
        fields={
            "footprint": PartField(
                value="Resistor_SMD:R_0603_1608Metric",
                visible="False",
            ),
            "datasheet": PartField(
                value="www.kicad.org",
                visible="False",
            ),
            "value": PartField(value="10k"),
            "reference": PartField(value="R"),
            "description": PartField(
                value="10 kOhms ±1% 0.1W, 1/10W Chip Resistor 0603 (1608 Metric) Moisture Resistant Thick Film",
                visible="False",
            ),
            "keywords": PartField(value="RES passive smd", visible="False"),
        },
    )
}

provider_parts = [
    {
        "part_id": "acme-res-10k",
        "display_name": "RC0603FR-0710KL",
        "symbol_name": "R",
        "library_name": "Device",
        "summary": "10 kOhms ±1% 0603 thick film resistor",
        "manufacturer": "Yageo",
        "mpn": "RC0603FR-0710KL",
        "license": "CC-BY-4.0",
        "auth_required": False,
        "available_assets": ["symbol", "footprint"],
    },
    {
        "part_id": "acme-res-1k",
        "display_name": "RC0603FR-071KL",
        "symbol_name": "R",
        "library_name": "Device",
        "summary": "1 kOhms ±1% 0603 thick film resistor",
        "manufacturer": "Yageo",
        "mpn": "RC0603FR-071KL",
        "license": "CC-BY-4.0",
        "auth_required": False,
        "available_assets": ["symbol"],
    },
]


def symbol_payload(symbol_name: str) -> str:
    return f"""(kicad_symbol_lib (version 20220914) (generator kicad_symbol_editor)
  (symbol "{symbol_name}" (in_bom yes) (on_board yes)
    (property "Reference" "R" (at 0 0 0)
      (effects (font (size 1.27 1.27)))
    )
    (property "Value" "{symbol_name}" (at 0 0 0)
      (effects (font (size 1.27 1.27)))
    )
    (property "Footprint" "" (at 0 0 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    (property "Datasheet" "" (at 0 0 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    (symbol "{symbol_name}_0_1"
      (rectangle (start -1.27 -1.27) (end 1.27 1.27)
        (stroke (width 0) (type default))
        (fill (type background))
      )
    )
    (symbol "{symbol_name}_1_1"
      (pin passive line (at -3.81 0 0) (length 2.54)
        (name "PIN" (effects (font (size 1.27 1.27))))
        (number "1" (effects (font (size 1.27 1.27))))
      )
    )
  )
)
"""


def asset_size_and_sha256(asset_name: str) -> tuple[int, str]:
    payload, _content_type = download_payloads[asset_name]
    payload_bytes = payload.encode("utf-8")
    return len(payload_bytes), hashlib.sha256(payload_bytes).hexdigest()


def manifest_payload(request: Request, part_id: str) -> Dict[str, object]:
    root = base_url(request)

    manifests = {
        "acme-res-10k": {
            "part_id": "acme-res-10k",
            "display_name": "RC0603FR-0710KL",
            "summary": "10 kOhms ±1% 0603 thick film resistor",
            "license": "CC-BY-4.0",
            "assets": [
                {
                    "asset_type": "symbol",
                    "name": "acme-res-10k.kicad_sym",
                    "target_library": "Device",
                    "target_name": "RC0603FR-0710KL",
                    "content_type": "application/x-kicad-symbol",
                    "size_bytes": asset_size_and_sha256("acme-res-10k.kicad_sym")[0],
                    "sha256": asset_size_and_sha256("acme-res-10k.kicad_sym")[1],
                    "download_url": f"{root}/downloads/acme-res-10k.kicad_sym",
                    "required": True,
                },
                {
                    "asset_type": "footprint",
                    "name": "R_0603.pretty",
                    "target_library": "Resistor_SMD",
                    "target_name": "R_0603_1608Metric",
                    "content_type": "application/x-kicad-footprint",
                    "size_bytes": 44,
                    "sha256": "8d8090740282c9ec23541a148af0ae57543e0da581e00e714e066dc4a1adefb0",
                    "download_url": f"{root}/downloads/R_0603.pretty",
                    "required": False,
                },
            ],
        },
        "acme-res-1k": {
            "part_id": "acme-res-1k",
            "display_name": "RC0603FR-071KL",
            "summary": "1 kOhms ±1% 0603 thick film resistor",
            "license": "CC-BY-4.0",
            "assets": [
                {
                    "asset_type": "symbol",
                    "name": "acme-res-1k.kicad_sym",
                    "target_library": "Device",
                    "target_name": "RC0603FR-071KL",
                    "content_type": "application/x-kicad-symbol",
                    "size_bytes": asset_size_and_sha256("acme-res-1k.kicad_sym")[0],
                    "sha256": asset_size_and_sha256("acme-res-1k.kicad_sym")[1],
                    "download_url": f"{root}/downloads/acme-res-1k.kicad_sym",
                    "required": True,
                }
            ],
        },
    }

    if part_id not in manifests:
        raise HTTPException(status_code=404, detail=f"Part {part_id} not found")

    return manifests[part_id]


download_payloads = {
    "acme-res-10k.kicad_sym": (
        symbol_payload("RC0603FR-0710KL"),
        "application/x-kicad-symbol",
    ),
    "acme-res-1k.kicad_sym": (
        symbol_payload("RC0603FR-071KL"),
        "application/x-kicad-symbol",
    ),
    "R_0603.pretty": (
        "(module \"R_0603_1608Metric\" (layer \"F.Cu\"))\n",
        "application/x-kicad-footprint",
    ),
    "preview.svg": (
        "<svg xmlns=\"http://www.w3.org/2000/svg\" width=\"96\" height=\"32\"><rect width=\"96\" height=\"32\" fill=\"#f5f1e8\"/><text x=\"8\" y=\"20\" font-size=\"12\">RC0603FR</text></svg>\n",
        "image/svg+xml",
    ),
}


def inline_component_bundle(request: Request, part_id: str) -> str:
    manifest = manifest_payload(request, part_id)
    entries = []

    for asset in manifest["assets"]:
        payload, _content_type = download_payloads[asset["name"]]
        entries.append(
            {
                "type": asset["asset_type"],
                "name": asset.get("target_name") or asset["name"],
                "content": base64.b64encode(payload.encode("utf-8")).decode("ascii"),
                "compression": "NONE",
            }
        )

    return base64.b64encode(json.dumps(entries).encode("utf-8")).decode("ascii")


@app.get("/")
async def root():
    return RedirectResponse(url="/app")


@app.get("/app")
async def provider_app(request: Request):
    manifests = {part["part_id"]: manifest_payload(request, part["part_id"]) for part in provider_parts}
    inline_payloads = {
        part_id: inline_component_bundle(request, part_id) for part_id in manifests.keys()
    }
    provider_session = request.cookies.get("kicad_provider_session") == "active"

    html = f"""
<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1" />
  <title>Embedded Provider Demo</title>
  <style>
    :root {{
      --paper: #f4efe2;
      --ink: #1e2430;
      --accent: #ae3f2f;
      --accent-2: #1f5c52;
      --line: #d3c7ab;
      --card: #fffaf0;
    }}
    * {{ box-sizing: border-box; }}
    body {{
      margin: 0;
      font-family: "Iowan Old Style", "Palatino Linotype", serif;
      background:
        radial-gradient(circle at top left, rgba(174,63,47,0.09), transparent 30%),
        linear-gradient(180deg, #f8f3e8, var(--paper));
      color: var(--ink);
    }}
    main {{ padding: 20px; }}
    .header {{ display: flex; gap: 16px; align-items: end; justify-content: space-between; margin-bottom: 20px; }}
    .title {{ max-width: 42rem; }}
    h1 {{ margin: 0; font-size: 2rem; letter-spacing: 0.03em; }}
    .sub {{ margin-top: 8px; font-size: 0.95rem; opacity: 0.8; }}
    .auth {{
      background: var(--card);
      border: 1px solid var(--line);
      border-radius: 14px;
      padding: 14px;
      min-width: 260px;
      box-shadow: 0 10px 30px rgba(30,36,48,0.08);
    }}
    .search {{
      width: 100%;
      padding: 12px 14px;
      border: 1px solid var(--line);
      border-radius: 999px;
      background: rgba(255,255,255,0.8);
      margin-bottom: 16px;
      font: inherit;
    }}
    .grid {{ display: grid; gap: 14px; }}
    .card {{
      background: var(--card);
      border: 1px solid var(--line);
      border-radius: 16px;
      padding: 16px;
      box-shadow: 0 10px 30px rgba(30,36,48,0.08);
    }}
    .card h2 {{ margin: 0 0 6px; font-size: 1.05rem; }}
    .meta {{ font-size: 0.85rem; opacity: 0.74; }}
    .actions {{ display: flex; flex-wrap: wrap; gap: 8px; margin-top: 12px; }}
    button {{
      border: 0;
      border-radius: 999px;
      padding: 10px 14px;
      font: inherit;
      cursor: pointer;
      color: white;
      background: var(--accent);
    }}
    button.alt {{ background: var(--accent-2); }}
    button.ghost {{
      background: transparent;
      color: var(--ink);
      border: 1px solid var(--line);
    }}
    pre {{
      margin: 18px 0 0;
      padding: 12px;
      border-radius: 12px;
      background: #231f20;
      color: #f3ead8;
      min-height: 100px;
      white-space: pre-wrap;
      font: 12px/1.5 ui-monospace, monospace;
    }}
  </style>
</head>
<body>
  <main>
    <section class="header">
      <div class="title">
        <h1>Embedded Provider Demo</h1>
        <div class="sub">Provider-owned UI in the KiCad panel. KiCad bridge handles auth, import, download, and place.</div>
      </div>
      <div class="auth">
        <div id="auth-state">Not connected to KiCad yet.</div>
        <div class="actions" style="margin-top:10px">
          <button id="signin">Sign In</button>
          <button id="signout" class="ghost">Sign Out</button>
          <button id="refresh" class="ghost">Refresh State</button>
        </div>
      </div>
    </section>

    <input id="query" class="search" type="search" placeholder="Filter demo parts" />
    <section id="parts" class="grid"></section>
    <pre id="log">Waiting for KiCad session...</pre>
  </main>
  <script>
    (() => {{
      const RPC_VERSION = 1;
      const PARTS = {json.dumps(provider_parts)};
      const MANIFESTS = {json.dumps(manifests)};
      const INLINE_BUNDLES = {json.dumps(inline_payloads)};
      const SERVER_SESSION_ACTIVE = {json.dumps(provider_session)};
      const waiters = new Map();
      let sessionId = null;
      let messageCounter = 0;
      let authenticated = false;

      const logEl = document.getElementById("log");
      const authEl = document.getElementById("auth-state");
      const partsEl = document.getElementById("parts");
      const queryEl = document.getElementById("query");

      function log(message) {{
        const stamp = new Date().toLocaleTimeString();
        logEl.textContent = `[${{stamp}}] ${{message}}\\n` + logEl.textContent;
      }}

      function updateAuthLabel() {{
        authEl.textContent = authenticated
          ? `Authenticated. Provider session cookie: ${{SERVER_SESSION_ACTIVE ? "yes" : "pending reload"}}`
          : "Signed out.";
      }}

      function postToKiCad(payload) {{
        if (window.webkit?.messageHandlers?.kicad) {{
          window.webkit.messageHandlers.kicad.postMessage(payload);
          return;
        }}
        if (window.chrome?.webview?.postMessage) {{
          window.chrome.webview.postMessage(payload);
          return;
        }}
        if (window.external?.invoke) {{
          window.external.invoke(payload);
          return;
        }}
        log("No KiCad bridge found.");
      }}

      function envelope(command, parameters = {{}}, data = "") {{
        if (!sessionId) throw new Error("No KiCad session yet.");
        return {{
          version: RPC_VERSION,
          session_id: sessionId,
          message_id: ++messageCounter,
          command,
          parameters,
          data
        }};
      }}

      function rpc(command, parameters = {{}}, data = "") {{
        const payload = envelope(command, parameters, data);
        return new Promise((resolve, reject) => {{
          const timer = window.setTimeout(() => {{
            waiters.delete(payload.message_id);
            reject(new Error("RPC timeout"));
          }}, 15000);

          waiters.set(payload.message_id, (response) => {{
            window.clearTimeout(timer);
            if (response.status === "ERROR") {{
              reject(new Error(response.error_message || response.error_code || "RPC error"));
              return;
            }}
            resolve(response);
          }});

          postToKiCad(JSON.stringify(payload));
        }});
      }}

      function handleIncoming(raw) {{
        let payload = raw;
        if (typeof payload === "string") {{
          try {{
            payload = JSON.parse(payload);
          }} catch (_err) {{
            return;
          }}
        }}

        if (!payload || typeof payload !== "object") return;

        if (payload.command === "NEW_SESSION" && payload.response_to === undefined) {{
          sessionId = payload.session_id;
          messageCounter = payload.message_id || 0;
          postToKiCad(JSON.stringify({{
            version: RPC_VERSION,
            session_id: sessionId,
            message_id: ++messageCounter,
            response_to: payload.message_id,
            command: "NEW_SESSION",
            status: "OK",
            parameters: {{
              server_name: "Embedded Provider Demo",
              server_version: "1.0.0"
            }}
          }}));
          log(`Session ready: ${{sessionId}}`);
          refreshState();
          return;
        }}

        if (payload.response_to !== undefined) {{
          const waiter = waiters.get(payload.response_to);
          if (waiter) {{
            waiters.delete(payload.response_to);
            waiter(payload);
          }}
        }}
      }}

      window.kiclient = window.kiclient || {{}};
      window.kiclient.postMessage = handleIncoming;

      async function refreshState() {{
        try {{
          const source = await rpc("GET_SOURCE_INFO");
          authenticated = !!source.parameters?.authenticated;
          updateAuthLabel();
        }} catch (err) {{
          log(`State refresh failed: ${{err.message}}`);
        }}
      }}

      async function signIn() {{
        try {{
          await rpc("REMOTE_LOGIN", {{ interactive: true }});
          log("Browser auth started. Complete sign-in in the system browser.");
        }} catch (err) {{
          log(`Sign-in failed: ${{err.message}}`);
        }}
      }}

      async function signOut() {{
        try {{
          await rpc("REMOTE_LOGIN", {{ sign_out: true }});
          authenticated = false;
          accessToken = null;
          updateAuthLabel();
          log("Signed out.");
        }} catch (err) {{
          log(`Sign-out failed: ${{err.message}}`);
        }}
      }}

      async function transfer(partId, command, inlineMode = false) {{
        const part = PARTS.find((item) => item.part_id === partId);
        try {{
          if (inlineMode) {{
            await rpc("DL_COMPONENT", {{
              compression: "NONE",
              library: part.library_name
            }}, INLINE_BUNDLES[partId]);
          }} else {{
            await rpc(command, {{
              part_id: part.part_id,
              display_name: part.display_name,
              symbol_name: part.symbol_name,
              library_name: part.library_name,
              assets: MANIFESTS[partId].assets
            }});
          }}
          log(`${{inlineMode ? "DL_COMPONENT inline" : command}} ok for ${{part.display_name}}`);
        }} catch (err) {{
          log(`${{inlineMode ? "DL_COMPONENT inline" : command}} failed for ${{part.display_name}}: ${{err.message}}`);
        }}
      }}

      function renderParts(filter = "") {{
        const q = filter.trim().toLowerCase();
        const visible = PARTS.filter((part) =>
          !q ||
          part.display_name.toLowerCase().includes(q) ||
          part.mpn.toLowerCase().includes(q) ||
          part.summary.toLowerCase().includes(q)
        );

        partsEl.innerHTML = visible.map((part) => `
          <article class="card">
            <h2>${{part.display_name}}</h2>
            <div class="meta">${{part.manufacturer}} | ${{part.mpn}}</div>
            <p>${{part.summary}}</p>
            <div class="actions">
              <button data-url="${{part.part_id}}">Download URL</button>
              <button data-inline="${{part.part_id}}" class="alt">Download Inline</button>
              <button data-place="${{part.part_id}}" class="ghost">Place</button>
            </div>
          </article>
        `).join("");

        partsEl.querySelectorAll("[data-url]").forEach((button) => {{
          button.addEventListener("click", () => transfer(button.dataset.url, "DL_COMPONENT", false));
        }});
        partsEl.querySelectorAll("[data-inline]").forEach((button) => {{
          button.addEventListener("click", () => transfer(button.dataset.inline, "DL_COMPONENT", true));
        }});
        partsEl.querySelectorAll("[data-place]").forEach((button) => {{
          button.addEventListener("click", () => transfer(button.dataset.place, "PLACE_COMPONENT", false));
        }});
      }}

      document.getElementById("signin").addEventListener("click", signIn);
      document.getElementById("signout").addEventListener("click", signOut);
      document.getElementById("refresh").addEventListener("click", refreshState);
      queryEl.addEventListener("input", () => renderParts(queryEl.value));

      renderParts();
      updateAuthLabel();
    }})();
  </script>
</body>
</html>
"""

    return Response(content=html, media_type="text/html")


@app.get("/.well-known/kicad-remote-provider")
async def provider_metadata(request: Request):
    root = base_url(request)
    return {
        "provider_name": "KiCad Local Test Provider",
        "provider_version": "1.0.0",
        "api_base_url": root,
        "panel_url": f"{root}/app",
        "session_bootstrap_url": f"{root}/session/bootstrap",
        "auth": {
            "type": "oauth2",
            "metadata_url": f"{root}/.well-known/oauth-authorization-server",
            "client_id": "kicad-desktop-test",
            "scopes": ["openid", "parts.read"],
        },
        "capabilities": {
            "web_ui_v1": True,
            "parts_v1": True,
            "direct_downloads_v1": True,
            "inline_payloads_v1": True,
        },
        "max_download_bytes": 1048576,
        "supported_asset_types": ["symbol", "footprint"],
        "parts": {
            "endpoint_template": "/v1/parts/{part_id}",
        },
        "documentation_url": f"{root}/docs",
        "allow_insecure_localhost": True,
    }


@app.post("/session/bootstrap")
async def session_bootstrap(
    request: Request,
    access_token: str = Form(...),
    next_url: str = Form("/app"),
):
    if access_token != "local-test-token":
        raise HTTPException(status_code=401, detail="Invalid access token")

    root = base_url(request)
    if not next_url.startswith(root):
        raise HTTPException(status_code=400, detail="next_url must stay on provider origin")

    response = RedirectResponse(url=next_url, status_code=303)
    response.set_cookie(
        "kicad_provider_session",
        "active",
        httponly=True,
        samesite="lax",
    )
    return response


@app.get("/.well-known/oauth-authorization-server")
async def oauth_metadata(request: Request):
    root = base_url(request)
    return {
        "issuer": root,
        "authorization_endpoint": f"{root}/oauth/authorize",
        "token_endpoint": f"{root}/oauth/token",
        "revocation_endpoint": f"{root}/oauth/revoke",
    }


@app.get("/oauth/authorize")
async def oauth_authorize(redirect_uri: str, state: str):
    separator = "&" if "?" in redirect_uri else "?"
    return RedirectResponse(
        url=f"{redirect_uri}{separator}code=local-test-code&state={state}"
    )


@app.post("/oauth/token")
async def oauth_token():
    return {
        "token_type": "Bearer",
        "access_token": "local-test-token",
        "refresh_token": "local-refresh-token",
        "expires_in": 3600,
        "scope": "openid parts.read",
    }


@app.post("/oauth/revoke")
async def oauth_revoke():
    return Response(status_code=200)


@app.get("/v1/parts/{part_id}")
async def remote_provider_manifest(request: Request, part_id: str):
    return manifest_payload(request, part_id)


@app.get("/downloads/{asset_name}")
async def download_asset(asset_name: str):
    if asset_name not in download_payloads:
        raise HTTPException(status_code=404, detail=f"Asset {asset_name} not found")

    body, content_type = download_payloads[asset_name]
    return PlainTextResponse(body, media_type=content_type)


@app.get("/v1/")
async def get_endpoints():
    return {"categories": "", "parts": ""}


@app.get("/v1/categories.json", response_model=List[Category])
async def get_categories():
    return categories


@app.get("/v1/parts/category/{category_id}.json", response_model=List[PartsChooser])
async def get_parts_by_category(category_id: str):
    if category_id not in parts_by_category:
        raise HTTPException(status_code=404, detail=f"Category {category_id} not found")

    return parts_by_category[category_id]


@app.get("/v1/parts/{part_id}.json", response_model=DetailedPart)
async def get_part_details(part_id: str):
    if part_id not in detailed_parts:
        raise HTTPException(status_code=404, detail=f"Part {part_id} not found")

    return detailed_parts[part_id]


def main():
    parser = argparse.ArgumentParser(description="Run the KiCad HTTP/provider test server")
    parser.add_argument("--host", default="127.0.0.1")
    parser.add_argument("--port", default=8000, type=int)
    args = parser.parse_args()

    uvicorn.run(app, host=args.host, port=args.port)


if __name__ == "__main__":
    main()



#  PadArray.py
#
#  Copyright 2014 john <john@johndev>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from __future__ import division

import math
import pcbnew

class PadMaker(object):
    """!
    Useful construction functions for common types of pads, providing
    sensible defaults for common pads.
    """

    def __init__(self, module):
        """!
        @param module: the module the pads will be part of
        """
        self.module = module

    def THPad(self, Vsize, Hsize, drill, shape=pcbnew.PAD_SHAPE_OVAL,
              rot_degree = 0):
        """!
        A basic through-hole pad of the given size and shape
        @param Vsize: the vertical size of the pad
        @param Hsize: the horizontal size of the pad
        @param drill: the drill diameter
        @param shape: the shape of the pad
        @param rot_degree: the pad rotation, in degrees
        """
        pad = pcbnew.PAD(self.module)
        pad.SetSize(pcbnew.VECTOR2I( int(Hsize), int(Vsize) ))
        pad.SetShape(shape)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_PTH)
        pad.SetLayerSet(pad.PTHMask())
        pad.SetDrillSize(pcbnew.VECTOR2I( int(drill), int(drill) ))
        pad.SetOrientation( pcbnew.EDA_ANGLE( rot_degree, pcbnew.DEGREES_T ) )

        return pad

    def THRoundPad(self, size, drill):
        """!
        A round though-hole pad. A shortcut for THPad()
        @param size: pad diameter
        @param drill: drill diameter
        """
        pad = self.THPad(size, size, drill, shape=pcbnew.PAD_SHAPE_CIRCLE)
        return pad

    def NPTHRoundPad(self, drill):
        """!
        A round non-plated though hole (NPTH)

        @param drill: the drill diameter (equals the NPTH diameter)
        """
        pad = pcbnew.PAD(self.module)
        pad.SetSize(pcbnew.VECTOR2I( int(drill), int(drill) ))
        pad.SetShape(pcbnew.PAD_SHAPE_CIRCLE)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_NPTH)
        pad.SetLayerSet(pad.UnplatedHoleMask())
        pad.SetDrillSize(pcbnew.VECTOR2I( int(drill), int(drill) ))
        return pad

    def SMDPad(self, Vsize, Hsize, shape=pcbnew.PAD_SHAPE_RECT, rot_degree=0):
        """
        Create a surface-mount pad of the given size and shape
        @param Vsize: the vertical size of the pad
        @param Hsize: the horizontal size of the pad
        @param shape: the shape of the pad
        @param rot_degree: the pad rotation, in degrees
        """
        pad = pcbnew.PAD(self.module)
        pad.SetSize(pcbnew.VECTOR2I( int(Hsize), int(Vsize) ) )
        pad.SetShape(shape)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_SMD)
        pad.SetLayerSet(pad.SMDMask())
        pad.SetOrientation( pcbnew.EDA_ANGLE( rot_degree, pcbnew.DEGREES_T ) )

        return pad

    def AperturePad(self, Vsize, Hsize, shape=pcbnew.PAD_SHAPE_RECT, rot_degree=0):
        """
        Create a aperture pad of the given size and shape, i.e. a smd pad shape
        on the solder paste and not on a copper
        layer
        @param Vsize: the vertical size of the aperture
        @param Hsize: the horizontal size of the aperture
        @param shape: the shape of the pad
        @param rot_degree: the pad rotation, in degrees
        """
        pad = pcbnew.PAD(self.module)
        pad.SetSize(pcbnew.VECTOR2I( int(Hsize), int(Vsize) ) )
        pad.SetShape(shape)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_SMD)
        pad.SetLayerSet(pad.ApertureMask())
        pad.SetOrientation( pcbnew.EDA_ANGLE( rot_degree, pcbnew.DEGREES_T ) )

        return pad

    def SMTRoundPad(self, size):
        """!
        A round surface-mount pad. A shortcut for SMDPad()
        @param size: pad diameter
        """
        pad = self.SMDPad(size, size, shape=pcbnew.PAD_SHAPE_CIRCLE)
        return pad


class PadArray(object):
    """!
    A class to assist in creating repetitive grids of pads

    Generally, PadArrays have an internal prototypical pad, and copy this
    for each pad in the array. They can also have a special pad for the
    first pad, and a custom function to name the pad.

    Generally, PadArray is used as a base class for more specific array
    types.
    """

    def __init__(self, pad):
        """!
        @param pad: the prototypical pad
        """
        self.firstPadNum = 1
        self.pinNames = None

        # this pad is more of a "context", we will use it as a source of
        # pad data, but not actually add it
        self.pad = pad
        self.firstPad = None

    def SetPinNames(self, pinNames):
        """!
        Set a name for all the pins. If given, this overrides the
        naming function.

        @param pinNames: the name to use for all pins
        """
        self.pinNames = pinNames

    def SetFirstPadType(self, firstPad):
        """!
        If the array has a different first pad, this is the pad that
        is used
        @param firstPad: the prototypical first pad
        """
        self.firstPad = firstPad

    def SetFirstPadInArray(self, fpNum):
        """!
        Set the numbering for the first pad in the array
        @param fpNum: the number for the first pad
        """
        self.firstPadNum = fpNum

    def AddPad(self, pad):
        """!
        Add a pad to the array, under the same footprint as the main
        prototype pad
        @param pad: pad to add
        """
        self.pad.GetParent().Add(pad)

    def GetPad(self, is_first_pad, pos):
        """!
        Get a pad in the array with the given position
        @param is_first_pad: use the special first pad if there is one
        @param pos: the pad position
        """
        if (self.firstPad and is_first_pad):
            pad = self.firstPad
        else:
            pad = self.pad

        # create a new pad with same characteristics
        pad = pad.Duplicate()
        pad.SetPosition(pos)

        return pad

    def GetName(self, *args, **kwargs):
        """!
        Get the pad name from the naming function, or the pre-set
        pinNames parameter (set with SetPinNames)
        """

        if self.pinNames is None:
            return self.NamingFunction(*args, **kwargs)

        return self.pinNames

    def NamingFunction(self, *args, **kwargs):
        """!
        Implement this as needed for each array type
        """
        raise NotImplementedError;


class PadGridArray(PadArray):
    """!
    A basic grid of pads
    """

    def __init__(self, pad, nx, ny, px, py, centre=pcbnew.VECTOR2I(0, 0)):
        """!
        @param pad: the prototypical pad of the array
        @param nx: number of pads in x-direction
        @param ny: number of pads in y-direction
        @param px: pitch in x-direction
        @param py: pitch in y-direction
        @param centre: array centre point
        """

        try:
            super().__init__(pad)
        except TypeError:
            super(PadGridArray, self).__init__(pad)

        self.nx = int(nx)
        self.ny = int(ny)
        self.px = px
        self.py = py
        self.centre = centre

    def AlphaNameFromNumber(self, n, aIndex=1,
                            alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
        """!
        Utility function to generate an alphabetical name:

        eg. 1 - A, 2 - B, 26 - AA, etc

        @param aIndex: index of 'A': 0 for 0 - A
        @param n: the pad index
        @param alphabet: set of allowable chars if not A-Z,
            e.g. ABCDEFGHJKLMNPRTUVWY for BGA
        """

        div, mod = divmod(n - aIndex, len(alphabet))
        alpha = alphabet[mod]

        if div > 0:
            return self.AlphaNameFromNumber(div, aIndex, alphabet) + alpha

        return alpha

    def NamingFunction(self, x, y):
        """!
        Implementation of the naming function: right to left, top-to-bottom

        @param x: the pad x index
        @param y: the pad y index
        """
        return self.firstPadNum + (self.nx * y + x)

    #relocate the pad and add it as many times as we need
    def AddPadsToModule(self, dc):
        """!
        Create the pads and add them to the module in the correct positions

        @param dc: the drawing context
        """

        pin1posX = self.centre.x - self.px * (self.nx - 1) / 2
        pin1posY = self.centre.y - self.py * (self.ny - 1) / 2

        for x in range(0, self.nx):
            posX = pin1posX + (x * self.px)

            for y in range(self.ny):
                posY = pin1posY + (self.py * y)
                pos = dc.TransformPoint(posX, posY)
                pad = self.GetPad(x == 0 and y == 0, pos)
                pad.SetName(self.GetName(x,y))
                self.AddPad(pad)


class EPADGridArray(PadGridArray):
    """!
    A pad grid array with a fixed name, used for things like thermal
    pads and via grids.
    """

    def NamingFunction(self, nx, ny):
        """!
        Simply return the firstPadNum
        @param nx: not used
        @param ny: not used
        """
        return self.firstPadNum


class PadZGridArray(PadArray):
    """!
    A staggered pin array
    """

    def __init__(self, pad, pad_count, line_count, line_pitch,
                 pad_pitch, centre=pcbnew.VECTOR2I(0, 0)):
        """!
        @param pad: the prototypical pad
        @param pad_count: total pad count
        @param line_count: number of staggered lines
        @param line_pitch: distance between lines
        @param pad_pitch: distance between pads in a line
        @param centre: array centre point
        """
        super(PadZGridArray, self).__init__(pad)

        self.pad_count = int(pad_count)
        self.line_count = int(line_count)
        self.line_pitch = line_pitch
        self.pad_pitch = pad_pitch
        self.centre = centre

    def NamingFunction(self, pad_pos):
        """!
        Naming just increased with pad index in array
        """
        return self.firstPadNum + pad_pos

    def AddPadsToModule(self, dc):
        """!
        Create the pads and add them to the module in the correct positions

        @param dc: the drawing context
        """

        pin1posX = self.centre.x - self.pad_pitch * (self.pad_count - 1) / 2
        pin1posY = self.centre.y + self.line_pitch * (self.line_count - 1) / 2
        line = 0

        for padnum in range(0, self.pad_count):
            posX = pin1posX + (padnum * self.pad_pitch)
            posY = pin1posY - (self.line_pitch * line)

            pos = dc.TransformPoint(posX, posY)
            pad = self.GetPad(padnum == 0, pos)
            pad.SetName(self.GetName(padnum))
            self.AddPad(pad)

            line += 1

            if line >= self.line_count:
                line = 0


class PadLineArray(PadGridArray):
    """!
    Shortcut cases for a single-row grid array. Can be used for
    constructing sections of larger footprints.
    """

    def __init__(self, pad, n, pitch, isVertical,
                 centre=pcbnew.VECTOR2I(0, 0)):
        """!
        @param pad: the prototypical pad
        @param n: number of pads in array
        @param pitch: distance between pad centres
        @param isVertical: horizontal or vertical array (can also use the
        drawing contexts transforms for more control)
        @param centre: array centre
        """

        if isVertical:
            super(PadLineArray, self).__init__(pad, 1, n, 0, pitch, centre)
        else:
            super(PadLineArray, self).__init__(pad, n, 1, pitch, 0, centre)


class PadCircleArray(PadArray):
    """!
    Circular pad array
    """

    def __init__(self, pad, n, r, angle_offset=0, centre=pcbnew.VECTOR2I(0, 0),
                 clockwise=True, padRotationEnable=False, padRotationOffset=0):
        """!
        @param pad: the prototypical pad
        @param n: number of pads in array
        @param r: the circle radius
        @param angle_offset: angle of the first pad
        @param centre: array centre point
        @param clockwise: array increases in a clockwise direction
        @param padRotationEnable: also rotate pads when placing
        @param padRotationOffset: rotation of first pad
        """

        super(PadCircleArray, self).__init__(pad)

        self.n = int(n)
        self.r = r
        self.angle_offset = angle_offset
        self.centre = centre
        self.clockwise = clockwise
        self.padRotationEnable = padRotationEnable
        self.padRotationOffset = padRotationOffset

    def NamingFunction(self, n):
        """!
        Naming around the circle, CW or CCW according to the clockwise flag
        """
        return str(self.firstPadNum + n)

    def AddPadsToModule(self, dc):
        """!
        Create the pads and add them to the module in the correct positions

        @param dc: the drawing context
        """

        for pin in range(0, self.n):
            angle = self.angle_offset + (360 / self.n) * pin

            if not self.clockwise:
                angle = -angle

            pos_x = math.sin(angle * math.pi / 180) * self.r
            pos_y = -math.cos(angle  * math.pi / 180) * self.r
            pos = dc.TransformPoint(pos_x, pos_y)
            pad = self.GetPad(pin == 0, pos)
            padAngle = self.padRotationOffset
            if self.padRotationEnable:
                padAngle -=angle
            pad.SetOrientation( pcbnew.EDA_ANGLE( padAngle, pcbnew.DEGREES_T ) )
            pad.SetName(self.GetName(pin))
            self.AddPad(pad)


class PadCustomArray(PadArray):
    """!
    Layout pads according to a custom array of [x,y] data
    """

    def __init__(self, pad, array):
        """!
        @param pad: the prototypical pad
        @param array: the position data array
        """
        super(PadCustomArray, self).__init__(pad)

        self.array = array

    def NamingFunction(self, n):
        """!
        Simple increment along the given array
        @param n: the pad index in the array
        """
        return str(self.firstPadNum + n)

    def AddPadsToModule(self, dc):
        """!
        Create the pads and add them to the module in the correct positions

        @param dc: the drawing context
        """

        for i in range(len(self.array)):
            pos = dc.TransformPoint(self.array[i][0], self.array[i][1])
            pad = self.GetPad(i == 0, pos)
            pad.SetName(self.GetName(i))
            self.AddPad(pad)



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from __future__ import division
import functools
import pcbnew as B
import FootprintWizardBase

'''
Created on Jan 16, 2015

@author: ejohns
Rewritten by LordBlick
'''
ptd = {
    '0': '000110100', '1': '100100001', '2': '001100001', '3': '101100000',
    '4': '000110001', '5': '100110000', '6': '001110000', '7': '000100101',
    '8': '100100100', '9': '001100100', 'A': '100001001', 'B': '001001001',
    'C': '101001000', 'D': '000011001', 'E': '100011000', 'F': '001011000',
    'G': '000001101', 'H': '100001100', 'I': '001001100', 'J': '000011100',
    'K': '100000011', 'L': '001000011', 'M': '101000010', 'N': '000010011',
    'O': '100010010', 'P': '001010010', 'Q': '000000111', 'R': '100000110',
    'S': '001000110', 'T': '000010110', 'U': '110000001', 'V': '011000001',
    'W': '111000000', 'X': '010010001', 'Y': '110010000', 'Z': '011010000',
    '-': '010000101', '.': '110000100', ' ': '011000100', '*': '010010100',
    '$': '010101000', '/': '010100010', '+': '010001010', '%': '000101010'}

class Uss39:
     def __init__(self, text):
        self.Text = self.makePrintable(text)

     __str__ = lambda self: self.Text
     makePrintable = lambda self, text: ''.join((c for c in text.upper() if c in ptd))

     def getBarCodePattern(self, text = None):
        text = text if not(text is None) else self.Text
        # Reformatted text with start and end characters
        return functools.reduce(lambda a1, a2: list(a1) + [0] + list(a2), [map(int, ptd[c]) for c in ("*%s*" % self.makePrintable(text))])

class Uss39Wizard(FootprintWizardBase.FootprintWizard):
    GetName = lambda self: 'BARCODE USS-39'
    GetDescription = lambda self: 'USS-39 Barcode'
    GetReferencePrefix = lambda self: 'BARCODE'
    GetValue = lambda self: self.module.Value().GetText()

    def GenerateParameterList(self):
        # Silkscreen parameters
        self.AddParam("Barcode", "Pixel Width", self.uMM, 0.20)
        self.AddParam("Barcode", "Height", self.uMM, 3.0)
        self.AddParam("Barcode", "Margin", self.uMM, 2.0)
        self.AddParam("Barcode", "Contents", self.uString, 'BARCODE')

        self.AddParam("Caption", "Enabled", self.uBool, True)
        self.AddParam("Caption", "Height", self.uMM, 1.2)
        self.AddParam("Caption", "Thickness", self.uMM, 0.12)

    def CheckParameters(self):

        # Reset constants
        self.CourtyardLineWidth = B.FromMM(0.05)
        # Set bar height to the greater of 6.35mm or 0.15*L
        # Set quiet width to 10*X
        # User-defined parameters
        # Create barcode object
        self.Barcode = Uss39('=' + str(self.parameters['Barcode']['Contents']))
        self.X = int(self.parameters['Barcode']['Pixel Width'])
        self.module.Value().SetText( str(self.Barcode) )
        self.C = len(str(self.Barcode))
        # Inter-character gap
        if self.X < 0.250:
            self.I = B.FromMM(3.15)
        else:
            self.I = (2 * self.X) if (2*self.X) > B.FromMM(1.35) else B.FromMM(1.35)
        # Wide to narrow ratio
        if self.X >= B.FromMM(0.508):
            self.N = B.FromMM(int((2.0+3.0)/2))
        else:
            self.N = B.FromMM(int((2.2+3.0)/2))
        self.H = self.parameters['Barcode']['Height']
        self.Q = (10 * self.X) if (10 * self.X) >  B.FromMM(6.35) else B.FromMM(6.35)
        self.L = self.I * (1 + self.C) + (self.C + 2) * (6 * self.X + 3 * self.N * self.X) + 2 * self.Q


    def __drawBar__(self, bit, x):
        offset = (bit + 1) * self.X
        return x + offset

    def __drawSpace__(self, bit, x):
        self.draw.SetLayer(B.F_SilkS)
        self.draw.SetLineThickness(self.X)
        self.draw.Line(x, 0, x, self.H)
        if (bit == 1):
            self.draw.Line(x + self.X, 0, x + self.X, self.H)
            self.draw.Line(x + self.X/2, 0, x + self.X/2, self.H)
            self.draw.Line(x, 0, x + self.X, 0)
            self.draw.Line(x, self.H, x + self.X, self.H)
        offset = (bit + 1) * self.X
        return x + offset

    def drawBars(self):
        x = 0
        bars = self.Barcode.getBarCodePattern()
        for index in range(0, len(bars), 2):
            # Draw bar
            barBit = bars[index]
            x = self.__drawBar__(barBit, x)
            # Draw space
            if index < len(bars)-1:
                spaceBit = bars[index + 1]
                x = self.__drawSpace__(spaceBit, x)
        return x

    def drawQuietZone(self, x0, y0, width, height):
        self.draw.SetLayer(B.F_SilkS)
        self.draw.SetLineThickness(self.X)

        for offset in range(0, int(self.Q), int(self.X/2)):
            xoffset = offset + self.X
            yoffset = offset + self.X/2
            self.draw.Line(x0 - xoffset, -yoffset, width + xoffset, -yoffset)
            self.draw.Line(x0 - xoffset, self.H+yoffset, width + xoffset, self.H+yoffset)
            self.draw.Line(x0 - xoffset, -yoffset, x0-xoffset, self.H+yoffset)
            self.draw.Line(width + xoffset, -yoffset, width+xoffset, self.H+yoffset)

    def BuildThisFootprint(self):
        # Draw bars
        x = self.drawBars()
        # Draw quiet zone
        self.drawQuietZone(0, 0, x, self.H)
        # Draw courtyard origin
        self.draw.SetLayer(B.F_CrtYd)
        self.draw.SetLineThickness(self.CourtyardLineWidth)
        ch_lim = B.FromMM(0.35)
        self.draw.Line(-ch_lim, 0, ch_lim, 0)
        self.draw.Line(0, -ch_lim, 0, ch_lim)
        self.draw.Circle(0, 0, B.FromMM(0.25))
        self.module.Value().SetLayer(B.F_Fab)

Uss39Wizard().register()


#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2023 Mark Roszko <mark.roszko@gmail.com>
# Copyright (C) 2023 KiCad Developers
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import utils
import cairosvg
import re
from pathlib import Path
import pytest
from typing import List

from conftest import KiTestFixture


@pytest.mark.parametrize("test_file,output_dir,compare_fn,cli_args",
                            [("cli/basic_test/basic_test.kicad_sch", "basic_test", "cli/basic_test/basic_test.svg", []),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test_nobg_bnw_nods", "cli/basic_test/basic_test_nobg_bnw_nods.svg", ["--no-background-color", "--exclude-drawing-sheet", "--black-and-white"])
                             ])
def test_sch_export_svg( kitest: KiTestFixture,
                         test_file: str,
                         output_dir: str,
                         compare_fn: str,
                         cli_args: List[str] ):
    input_file = kitest.get_data_file_path( test_file )

    output_path =  kitest.get_output_path( "cli/{}/".format( output_dir ) )

    command = [utils.kicad_cli(), "sch", "export", "svg"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_path ) )
    command.append( str(input_file) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0
    assert stderr == ''
    assert stdout is not None

    stdout_regex = re.search("Plotted to '(.+)'", stdout)
    assert stdout_regex

    # now try and manipulate the extracted path
    output_svg_path = Path( stdout_regex.group(1) )
    assert output_svg_path.exists()

    kitest.add_attachment( output_svg_path )

    png_converted_from_svg_path = output_svg_path.with_suffix( '.png' )

    cairosvg.svg2png( url=str( output_svg_path ), write_to=str( png_converted_from_svg_path ), dpi=1200 )

    compare_file_path = kitest.get_data_file_path( compare_fn )
    compare_stem = f"orig_{output_dir}"
    compare_png_converted_from_svg_path = output_svg_path.with_suffix( '.png' ).with_stem(compare_stem)
    cairosvg.svg2png( url=str( compare_file_path ), write_to=str( compare_png_converted_from_svg_path ), dpi=1200 )

    assert utils.images_are_equal( png_converted_from_svg_path, compare_png_converted_from_svg_path,
                                   diff_handler=kitest.add_attachment )


@pytest.mark.parametrize("test_file,output_fn,line_skip_count,skip_compare,cli_args",
                            [("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.kicadsexpr", 5, True, []),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.kicadsexpr", 5, True,["--format=kicadsexpr"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.kicadxml", 6, True,["--format=kicadxml"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.cadstar", 3, False, ["--format=cadstar"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.orcadpcb2", 1, False, ["--format=orcadpcb2"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.pads", 0, False, ["--format=pads"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.netlist.allegro", 3, False, ["--format=allegro"]),
                             ])
def test_sch_export_netlist( kitest,
                             test_file: str,
                             output_fn: str,
                             line_skip_count: int,
                             skip_compare: bool,
                             cli_args: List[str] ):
    input_file = kitest.get_data_file_path( test_file )
    compare_filepath = kitest.get_data_file_path( "cli/basic_test/{}".format( output_fn ) )

    output_filepath =  kitest.get_output_path( "cli/" ).joinpath( output_fn )

    command = [utils.kicad_cli(), "sch", "export", "netlist"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str(input_file) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0
    assert stderr == ''

    # some of our netlist formats are not cross platform so skip for now
    if not skip_compare:
        assert utils.textdiff_files( compare_filepath, output_filepath, line_skip_count )

    kitest.add_attachment( output_filepath )


@pytest.mark.parametrize("test_file,output_fn,cli_args",
                            [("cli/basic_test/basic_test.kicad_sch", "basic_test.pdf", []),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bnw.nods.nobg.pdf", ["--black-and-white","--exclude-drawing-sheet","--no-background-color"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.pone.pdf", ["--pages", "1"])
                             ])
def test_sch_export_pdf( kitest,
                         test_file: str,
                         output_fn: str,
                         cli_args: List[str] ):
    input_file = kitest.get_data_file_path( test_file )

    output_filepath =  kitest.get_output_path( "cli/" ).joinpath( output_fn )

    command = [utils.kicad_cli(), "sch", "export", "pdf"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0
    assert stderr == ''

    kitest.add_attachment( output_filepath )


@pytest.mark.parametrize("test_file,output_fn,compare_fn,line_skip_count,cli_args",
                            [("cli/variants/variants.kicad_sch", "variants_default.bom.csv", "cli/variants/variants_default.bom.csv", 0,
                              ["--exclude-dnp", "--fields", "Reference,Value", "--labels", "Refs,Value"]),
                             ("cli/variants/variants.kicad_sch", "variants_v1.bom.csv", "cli/variants/variants_v1.bom.csv", 0,
                              ["--variant", "Variant 1", "--exclude-dnp", "--fields", "Reference,Value", "--labels", "Refs,Value"]),
                             ("cli/variants/variants.kicad_sch", "variants_v2.bom.csv", "cli/variants/variants_v2.bom.csv", 0,
                              ["--variant", "Variant2", "--exclude-dnp", "--fields", "Reference,Value", "--labels", "Refs,Value"]),
                             ])
def test_sch_export_bom_variants( kitest: KiTestFixture,
                         test_file: str,
                         output_fn: str,
                         compare_fn: str,
                         line_skip_count: int,
                         cli_args: List[str] ):
    """Test BOM export with variant support and DNP exclusion"""
    input_file = kitest.get_data_file_path( test_file )
    compare_filepath = kitest.get_data_file_path( compare_fn )

    output_filepath =  kitest.get_output_path( "cli/" ).joinpath( output_fn )

    command = [utils.kicad_cli(), "sch", "export", "bom"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0
    assert stderr == ''

    assert utils.textdiff_files( compare_filepath, output_filepath, line_skip_count )

    kitest.add_attachment( output_filepath )


def test_sch_export_bom_multi_variant_requires_placeholder( kitest: KiTestFixture ):
    """Test that multiple variants require ${VARIANT} in output path"""
    input_file = kitest.get_data_file_path( "cli/variants/variants.kicad_sch" )

    output_filepath = kitest.get_output_path( "cli/" ).joinpath( "multi_variant_fail.csv" )

    command = [utils.kicad_cli(), "sch", "export", "bom"]
    command.extend( ["--variant", "Variant 1", "--variant", "Variant2"] )
    command.extend( ["--exclude-dnp", "--fields", "Reference,Value", "--labels", "Refs,Value"] )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 1
    assert "VARIANT" in stderr


def test_sch_export_bom_multi_variant_with_placeholder( kitest: KiTestFixture ):
    """Test BOM export with multiple variants using ${VARIANT} placeholder"""
    input_file = kitest.get_data_file_path( "cli/variants/variants.kicad_sch" )

    output_dir = kitest.get_output_path( "cli/" )
    output_pattern = str( output_dir.joinpath( "bom_${VARIANT}.csv" ) )

    command = [utils.kicad_cli(), "sch", "export", "bom"]
    command.extend( ["--variant", "Variant 1", "--variant", "Variant2"] )
    command.extend( ["--exclude-dnp", "--fields", "Reference,Value", "--labels", "Refs,Value"] )
    command.append( "-o" )
    command.append( output_pattern )
    command.append( str(input_file) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0
    assert stderr == ''

    v1_path = output_dir.joinpath( "bom_Variant 1.csv" )
    v2_path = output_dir.joinpath( "bom_Variant2.csv" )

    assert v1_path.exists(), f"Expected output file {v1_path} not found"
    assert v2_path.exists(), f"Expected output file {v2_path} not found"

    v1_compare = kitest.get_data_file_path( "cli/variants/variants_v1.bom.csv" )
    v2_compare = kitest.get_data_file_path( "cli/variants/variants_v2.bom.csv" )

    assert utils.textdiff_files( v1_compare, v1_path, 0 )
    assert utils.textdiff_files( v2_compare, v2_path, 0 )

    kitest.add_attachment( v1_path )
    kitest.add_attachment( v2_path )


@pytest.mark.parametrize("test_file,output_fn,line_skip_count,cli_args",
                            [("cli/basic_test/basic_test.kicad_sch", "basic_test.pythonbom", 6, [])
                             ])
def test_sch_export_pythonbom( kitest: KiTestFixture,
                         test_file: str,
                         output_fn: str,
                         line_skip_count: int,
                         cli_args: List[str] ):
    input_file = kitest.get_data_file_path( test_file )
    compare_filepath = kitest.get_data_file_path( "cli/basic_test/{}".format( output_fn ) )

    output_filepath =  kitest.get_output_path( "cli/" ).joinpath( output_fn )

    command = [utils.kicad_cli(), "sch", "export", "python-bom"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0
    assert stderr == ''

    # pythonbom is not currently crossplatform (platform specific paths) to enable diffs

    kitest.add_attachment( output_filepath )



@pytest.mark.parametrize("test_file,output_fn,line_skip_count,skip_compare,expected_exit_code,cli_args",
                            [("cli/basic_test/basic_test.kicad_sch", "basic_test.erc.rpt", 1, False, 0, []),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.erc.rpt", 1, False, 0, ["--format=report"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.erc.json", 5, False,0, ["--format=json"]),
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.erc.unitsin.rpt", 1, False, 0, ["--format=report", "--units=in"]),
                             ])
def test_sch_export_erc( kitest: KiTestFixture,
                             test_file: str,
                             output_fn: str,
                             line_skip_count: int,
                             skip_compare: bool,
                             expected_exit_code: int,
                             cli_args: List[str] ):
    input_file = kitest.get_data_file_path( test_file )
    compare_filepath = kitest.get_data_file_path( "cli/basic_test/{}".format( output_fn ) )

    output_filepath =  kitest.get_output_path( "cli/" ).joinpath( output_fn )

    command = [utils.kicad_cli(), "sch", "erc"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == expected_exit_code

    # some of our netlist formats are not cross platform so skip for now
    if not skip_compare:
        assert utils.textdiff_files( compare_filepath, output_filepath, line_skip_count )

    kitest.add_attachment( output_filepath )


@pytest.mark.parametrize("test_file,output_fn,expected_headers,cli_args",
                            [
                             # Default fields include QUANTITY and DNP
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bom_default.csv",
                              ["Refs", "Value", "Footprint", "Qty", "DNP"], []),
                             # Explicit fields with ${QUANTITY} (canonical form)
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bom_quantity.csv",
                              ["Refs", "Value", "Qty"],
                              ["--fields", "${QUANTITY},Reference,Value", "--labels", "Qty,Refs,Value"]),
                             # Explicit fields with ${ITEM_NUMBER} (canonical form)
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bom_item_number.csv",
                              ["#", "Refs", "Value"],
                              ["--fields", "${ITEM_NUMBER},Reference,Value", "--labels", "#,Refs,Value"]),
                             # Bare field names without ${} delimiters (shell-safe form)
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bom_bare_names.csv",
                              ["Refs", "Value", "Qty", "DNP"],
                              ["--fields", "Reference,Value,QUANTITY,DNP", "--labels", "Refs,Value,Qty,DNP"]),
                             # Bare ITEM_NUMBER without ${} delimiters
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bom_bare_item.csv",
                              ["#", "Refs", "Value"],
                              ["--fields", "ITEM_NUMBER,Reference,Value", "--labels", "#,Refs,Value"]),
                             # Bare EXCLUDE_FROM_BOM
                             ("cli/basic_test/basic_test.kicad_sch", "basic_test.bom_bare_excl.csv",
                              ["Refs", "Value", "Excl BOM"],
                              ["--fields", "Reference,Value,EXCLUDE_FROM_BOM",
                               "--labels", "Refs,Value,Excl BOM"]),
                             ])
def test_sch_export_bom( kitest,
                         test_file: str,
                         output_fn: str,
                         expected_headers: List[str],
                         cli_args: List[str] ):
    """Test BOM export with various field configurations, including virtual fields like QUANTITY."""
    input_file = kitest.get_data_file_path( test_file )

    output_filepath = kitest.get_output_path( "cli/" ).joinpath( output_fn )

    command = [utils.kicad_cli(), "sch", "export", "bom"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_filepath ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    assert exitcode == 0, f"BOM export failed with exit code {exitcode}: {stderr}"
    assert output_filepath.exists(), f"Output file not created: {output_filepath}"

    # Read the BOM file and verify headers
    with open( output_filepath, 'r' ) as f:
        first_line = f.readline().strip()

    # Parse the CSV header (removing quotes)
    actual_headers = [h.strip().strip('"') for h in first_line.split(',')]

    for expected in expected_headers:
        assert expected in actual_headers, f"Expected header '{expected}' not found in BOM output. Got: {actual_headers}"

    kitest.add_attachment( output_filepath )



'''
    A python script example to create plot files to build a board:
    Gerber files
    Drill files
    Map dril files

    Important note:
        this python script does not plot frame references (page layout).
        the reason is it is not yet possible from a python script because plotting
        plot frame references needs loading the corresponding page layout file
        (.wks file) or the default template.

        This info (the page layout template) is not stored in the board, and therefore
        not available.

        Do not try to change SetPlotFrameRef(False) to SetPlotFrameRef(true)
        the result is the pcbnew lib will crash if you try to plot
        the unknown frame references template.

        Anyway, in gerber and drill files the page layout is not plot
'''

import sys
import os
from pcbnew import *

# A helper function to convert a UTF8 string for python2 or python3
def fromUTF8Text( afilename ):
    if sys.version_info < (3, 0):
        return afilename.encode()
    else:
        return afilename

filename=sys.argv[1]

board = LoadBoard(filename)

plotDir = "plot/"

#prepare the gerber job file
gen_job_file=True

pctl = PLOT_CONTROLLER(board)

popt = pctl.GetPlotOptions()

popt.SetOutputDirectory(plotDir)

# Set some important plot options (see pcb_plot_params.h):
popt.SetPlotFrameRef(False)     #do not change it
popt.SetSketchPadLineWidth(FromMM(0.1))

popt.SetAutoScale(False)        #do not change it
popt.SetScale(1)                #do not change it
popt.SetMirror(False)
popt.SetUseGerberAttributes(True)
popt.SetIncludeGerberNetlistInfo(True)
popt.SetCreateGerberJobFile(gen_job_file)
popt.SetUseGerberProtelExtensions(False)
popt.SetUseAuxOrigin(True)

# This by gerbers only
popt.SetSubtractMaskFromSilk(False)
# Disable plot pad holes
popt.SetDrillMarksType( DRILL_MARKS_NO_DRILL_SHAPE );
# Skip plot pad NPTH when possible: when drill size and shape == pad size and shape
# usually sel to True for copper layers
popt.SetSkipPlotNPTH_Pads( False );


#prepare the gerber job file
jobfile_writer = GERBER_JOBFILE_WRITER( board )

# Once the defaults are set it become pretty easy...
# I have a Turing-complete programming language here: I'll use it...
# param 0 is a string added to the file base name to identify the drawing
# param 1 is the layer ID
# param 2 is a comment
plot_plan = [
    ( "CuTop", F_Cu, "Top layer" ),
    ( "In1Top", In1_Cu, "In1 layer" ),
    ( "CuBottom", B_Cu, "Bottom layer" ),
    ( "PasteBottom", B_Paste, "Paste Bottom" ),
    ( "PasteTop", F_Paste, "Paste top" ),
    ( "SilkTop", F_SilkS, "Silk top" ),
    ( "SilkBottom", B_SilkS, "Silk top" ),
    ( "MaskBottom", B_Mask, "Mask bottom" ),
    ( "MaskTop", F_Mask, "Mask top" ),
    ( "EdgeCuts", Edge_Cuts, "Edges" ),
]


# In Gerber format, Set layer before calling OpenPlotfile is mandatory to generate
# the right Gerber file header.
for layer_info in plot_plan:
    if layer_info[1] <= B_Cu:
        popt.SetSkipPlotNPTH_Pads( True )
    else:
        popt.SetSkipPlotNPTH_Pads( False )

    pctl.SetLayer(layer_info[1])
    pctl.OpenPlotfile(layer_info[0], PLOT_FORMAT_GERBER, layer_info[2])

    print( 'plot %s' % fromUTF8Text( pctl.GetPlotFileName() ) )

    if gen_job_file == True:
        jobfile_writer.AddGbrFile( layer_info[1], os.path.basename(pctl.GetPlotFileName()) );
    if pctl.PlotLayer() == False:
        print( "plot error" )

#generate internal copper layers, if any
lyrcnt = board.GetCopperLayerCount();

for innerlyr in range ( 1, lyrcnt-1 ):
    popt.SetSkipPlotNPTH_Pads( True );
    pctl.SetLayer(innerlyr)
    lyrname = 'inner%s' % innerlyr
    pctl.OpenPlotfile(lyrname, PLOT_FORMAT_GERBER, "inner")

    print( "plot %s" % fromUTF8Text( pctl.GetPlotFileName() ) )

    if pctl.PlotLayer() == False:
        print( "plot error" )


# At the end you have to close the last plot, otherwise you don't know when
# the object will be recycled!
pctl.ClosePlot()

# Fabricators need drill files.
# sometimes a drill map file is asked (for verification purpose)
drlwriter = EXCELLON_WRITER( board )
drlwriter.SetMapFileFormat( PLOT_FORMAT_PDF )

mirror = False
minimalHeader = False
offset = VECTOR2I(0,0)
# False to generate 2 separate drill files (one for plated holes, one for non plated holes)
# True to generate only one drill file
mergeNPTH = False
drlwriter.SetOptions( mirror, minimalHeader, offset, mergeNPTH )

metricFmt = True
drlwriter.SetFormat( metricFmt )

genDrl = True
genMap = True
print( 'create drill and map files in %s' % fromUTF8Text( pctl.GetPlotFileName() ) )
drlwriter.CreateDrillandMapFilesSet( pctl.GetPlotDirName(), genDrl, genMap );

# One can create a text file to report drill statistics
rptfn = pctl.GetPlotDirName() + 'drill_report.rpt'
print( 'report: %s' % fromUTF8Text( rptfn ) )
drlwriter.GenDrillReportFile( rptfn );

if gen_job_file == True:
    #job_fn=os.path.splitext(pctl.GetPlotFileName())[0] + '.gbrjob'
    job_fn=os.path.dirname(pctl.GetPlotFileName()) + '/' + os.path.basename(filename)
    job_fn=os.path.splitext(job_fn)[0] + '.gbrjob'

    print( 'create job file %s ' % fromUTF8Text( job_fn ) )

    jobfile_writer.CreateJobFile( job_fn )



#!/usr/bin/env python
# Created for KiCad project by Miguel
# Some modifications by Edwin
# GPL2

import subprocess
import os
import difflib


# class for checking and uncrustifying files
# defaults to cpp,cxx,h,hpp and c files
class coding_checker(object):
    file_filter = ["cpp", "cxx", "h", "hpp", "c"]

    # Function to call uncrustify, it returns the re-formatted code and
    # any errors
    #

    def uncrustify_file(self, filename=None):
        try:
            args = ("uncrustify", "-c", "uncrustify.cfg", "-f", filename)
            popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            popen.wait()
            return [popen.stdout.readlines(), popen.stderr.read()]
        except OSError as e:
            print "System returned : {e}\nCould not run uncrustify. Is it installed?".format(e=e.strerror)
            return [None, None]

    # This function runs bzr, and gets the list of modified files
    def bzr_modified(self):
        modified_files = []
        args = ("bzr", "status")
        try:
            popen = subprocess.Popen(args, stdout=subprocess.PIPE)
            popen.wait()
            output = popen.stdout.readlines()
        except OSError as e:
            print "System returned : {e}\nCould not run bzr. Is it installed?".format(e=e.strerror)
            return None

        in_modifieds = False
        for line in output:
            line = line.rstrip("\r\n")
            if line.endswith(":"):
                in_modifieds = False
            if line.startswith("modified:"):
                in_modifieds = True
                continue
            if line.startswith("added:"):
                in_modifieds = True
                continue

            if in_modifieds:
                modified_files.append(line.lstrip("\t ").rstrip("\t "))

        return modified_files


    def extension(self, filename):
        return os.path.splitext(filename)[1][1:].strip().lower()


    def read_file(self, filename):
        f = open(filename, 'r')
        data = f.readlines()
        f.close()
        return data


    def ask_user(self, filename):
        msg = 'Shall I clean %s ?' % filename
        return raw_input("%s (y/N/E) " % msg).lower()


    def main(self):
        # make list of modified file names
        modified_files = self.bzr_modified()

        if not modified_files:
            print "No modified files\n"
        else:

            for filename in modified_files:
                if self.extension(filename) in self.file_filter:
                    self.compare_and_suggest(filename)

    def compare_and_suggest(self,filename):
        # if it is a 'c' file try to uncrustify
        [uncrustified, errors] = self.uncrustify_file(filename)

        if not (uncrustified and errors):
            print "Program end"
            # problem in uncrustify
            return

        original = self.read_file(filename)

        if len(errors.split("\n")) > 2:
            print "There was a problem processing " + filename + ":" + errors
            return

        if uncrustified == original:
            print filename + " looks perfect!, well done!"
        else:
            print "Suggestions for: " + filename

            diff = difflib.unified_diff(original, uncrustified, filename, filename + ".uncrustified")

            for line in diff:
                print line.rstrip("\r\n")
            print ""
            reply = self.ask_user(filename)

            if reply in ["y", "yes"]:
                f = open(filename, 'w')
                for line in uncrustified:
                    f.write(line)
                f.close()
                print filename + " UPDATED"

            if reply in ["e", "ed", "edit"]:
                os.system("$EDITOR " + filename)

            print ""


if __name__ == '__main__':
    print "This program tries to do 2 things\n" \
          "1) call bzr to find changed files (related to the KiCad project)\n" \
          "2) call uncrustify on the changed files (to make the files comply with coding standards)\n"

    cc = coding_checker()
    cc.main()



# A basic mkdocstrings handler for {fmt}.
# Copyright (c) 2012 - present, Victor Zverovich
# https://github.com/fmtlib/fmt/blob/master/LICENSE

import os
import xml.etree.ElementTree as ElementTree
from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError, Popen
from typing import Any, List, Mapping, Optional

from mkdocstrings.handlers.base import BaseHandler


class Definition:
    """A definition extracted by Doxygen."""

    def __init__(self, name: str, kind: Optional[str] = None,
                 node: Optional[ElementTree.Element] = None,
                 is_member: bool = False):
        self.name = name
        self.kind = kind if kind is not None else node.get('kind')
        self.desc = None
        self.id = name if not is_member else None
        self.members = None
        self.params = None
        self.template_params = None
        self.trailing_return_type = None
        self.type = None


# A map from Doxygen to HTML tags.
tag_map = {
    'bold': 'b',
    'emphasis': 'em',
    'computeroutput': 'code',
    'para': 'p',
    'programlisting': 'pre',
    'verbatim': 'pre'
}

# A map from Doxygen tags to text.
tag_text_map = {
    'codeline': '',
    'highlight': '',
    'sp': ' '
}


def escape_html(s: str) -> str:
    return s.replace("<", "&lt;")


def doxyxml2html(nodes: List[ElementTree.Element]):
    out = ''
    for n in nodes:
        tag = tag_map.get(n.tag)
        if not tag:
            out += tag_text_map[n.tag]
        out += '<' + tag + '>' if tag else ''
        out += '<code class="language-cpp">' if tag == 'pre' else ''
        if n.text:
            out += escape_html(n.text)
        out += doxyxml2html(list(n))
        out += '</code>' if tag == 'pre' else ''
        out += '</' + tag + '>' if tag else ''
        if n.tail:
            out += n.tail
    return out


def convert_template_params(node: ElementTree.Element) -> Optional[List[Definition]]:
    template_param_list = node.find('templateparamlist')
    if template_param_list is None:
        return None
    params = []
    for param_node in template_param_list.findall('param'):
        name = param_node.find('declname')
        param = Definition(name.text if name is not None else '', 'param')
        param.type = param_node.find('type').text
        params.append(param)
    return params


def get_description(node: ElementTree.Element) -> List[ElementTree.Element]:
    return node.findall('briefdescription/para') + \
        node.findall('detaileddescription/para')


def normalize_type(type_: str) -> str:
    type_ = type_.replace('< ', '<').replace(' >', '>')
    return type_.replace(' &', '&').replace(' *', '*')


def convert_type(type_: ElementTree.Element) -> Optional[str]:
    if type_ is None:
        return None
    result = type_.text if type_.text else ''
    for ref in type_:
        result += ref.text
        if ref.tail:
            result += ref.tail
    result += type_.tail.strip()
    return normalize_type(result)


def convert_params(func: ElementTree.Element) -> List[Definition]:
    params = []
    for p in func.findall('param'):
        d = Definition(p.find('declname').text, 'param')
        d.type = convert_type(p.find('type'))
        params.append(d)
    return params


def convert_return_type(d: Definition, node: ElementTree.Element) -> None:
    d.trailing_return_type = None
    if d.type == 'auto' or d.type == 'constexpr auto':
        parts = node.find('argsstring').text.split(' -> ')
        if len(parts) > 1:
            d.trailing_return_type = normalize_type(parts[1])


def render_param(param: Definition) -> str:
    return param.type + (f'&nbsp;{param.name}' if len(param.name) > 0 else '')


def render_decl(d: Definition) -> str:
    text = ''
    if d.id is not None:
        text += f'<a id="{d.id}">\n'
    text += '<pre><code class="language-cpp decl">'

    text += '<div>'
    if d.template_params is not None:
        text += 'template &lt;'
        text += ', '.join([render_param(p) for p in d.template_params])
        text += '&gt;\n'
    text += '</div>'

    text += '<div>'
    end = ';'
    if d.kind == 'function' or d.kind == 'variable':
        text += d.type + ' ' if len(d.type) > 0 else ''
    elif d.kind == 'typedef':
        text += 'using '
    elif d.kind == 'define':
        end = ''
    else:
        text += d.kind + ' '
    text += d.name

    if d.params is not None:
        params = ', '.join([
            (p.type + ' ' if p.type else '') + p.name for p in d.params])
        text += '(' + escape_html(params) + ')'
        if d.trailing_return_type:
            text += ' -&NoBreak;>&nbsp;' + escape_html(d.trailing_return_type)
    elif d.kind == 'typedef':
        text += ' = ' + escape_html(d.type)

    text += end
    text += '</div>'
    text += '</code></pre>\n'
    if d.id is not None:
        text += f'</a>\n'
    return text


class CxxHandler(BaseHandler):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(handler='cxx', **kwargs)

        headers = [
            'args.h', 'base.h', 'chrono.h', 'color.h', 'compile.h', 'format.h',
            'os.h', 'ostream.h', 'printf.h', 'ranges.h', 'std.h', 'xchar.h'
        ]

        # Run doxygen.
        cmd = ['doxygen', '-']
        support_dir = Path(__file__).parents[3]
        top_dir = os.path.dirname(support_dir)
        include_dir = os.path.join(top_dir, 'include', 'fmt')
        self._ns2doxyxml = {}
        build_dir = os.path.join(top_dir, 'build')
        os.makedirs(build_dir, exist_ok=True)
        self._doxyxml_dir = os.path.join(build_dir, 'doxyxml')
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
        _, _ = p.communicate(input=r'''
            PROJECT_NAME     = fmt
            GENERATE_XML     = YES
            GENERATE_LATEX   = NO
            GENERATE_HTML    = NO
            INPUT            = {0}
            XML_OUTPUT       = {1}
            QUIET            = YES
            AUTOLINK_SUPPORT = NO
            MACRO_EXPANSION  = YES
            PREDEFINED       = _WIN32=1 \
                               __linux__=1 \
                               FMT_ENABLE_IF(...)= \
                               FMT_USE_USER_LITERALS=1 \
                               FMT_USE_ALIAS_TEMPLATES=1 \
                               FMT_USE_NONTYPE_TEMPLATE_ARGS=1 \
                               FMT_API= \
                               "FMT_BEGIN_NAMESPACE=namespace fmt {{" \
                               "FMT_END_NAMESPACE=}}" \
                               "FMT_DOC=1"
            '''.format(
                ' '.join([os.path.join(include_dir, h) for h in headers]),
                self._doxyxml_dir).encode('utf-8'))
        if p.returncode != 0:
            raise CalledProcessError(p.returncode, cmd)

        # Merge all file-level XMLs into one to simplify search.
        self._file_doxyxml = None
        for h in headers:
            filename = h.replace(".h", "_8h.xml")
            with open(os.path.join(self._doxyxml_dir, filename)) as f:
                doxyxml = ElementTree.parse(f)
                if self._file_doxyxml is None:
                    self._file_doxyxml = doxyxml
                    continue
                root = self._file_doxyxml.getroot()
                for node in doxyxml.getroot():
                    root.append(node)

    def collect_compound(self, identifier: str,
                         cls: List[ElementTree.Element]) -> Definition:
        """Collect a compound definition such as a struct."""
        path = os.path.join(self._doxyxml_dir, cls[0].get('refid') + '.xml')
        with open(path) as f:
            xml = ElementTree.parse(f)
            node = xml.find('compounddef')
            d = Definition(identifier, node=node)
            d.template_params = convert_template_params(node)
            d.desc = get_description(node)
            d.members = []
            for m in \
                    node.findall('sectiondef[@kind="public-attrib"]/memberdef') + \
                    node.findall('sectiondef[@kind="public-func"]/memberdef'):
                name = m.find('name').text
                # Doxygen incorrectly classifies members of private unnamed unions as
                # public members of the containing class.
                if name.endswith('_'):
                    continue
                desc = get_description(m)
                if len(desc) == 0:
                    continue
                kind = m.get('kind')
                member = Definition(name if name else '', kind=kind, is_member=True)
                type_text = m.find('type').text
                member.type = type_text if type_text else ''
                if kind == 'function':
                    member.params = convert_params(m)
                    convert_return_type(member, m)
                member.template_params = None
                member.desc = desc
                d.members.append(member)
            return d

    def collect(self, identifier: str, _config: Mapping[str, Any]) -> Definition:
        qual_name = 'fmt::' + identifier

        param_str = None
        paren = qual_name.find('(')
        if paren > 0:
            qual_name, param_str = qual_name[:paren], qual_name[paren + 1:-1]

        colons = qual_name.rfind('::')
        namespace, name = qual_name[:colons], qual_name[colons + 2:]

        # Load XML.
        doxyxml = self._ns2doxyxml.get(namespace)
        if doxyxml is None:
            path = f'namespace{namespace.replace("::", "_1_1")}.xml'
            with open(os.path.join(self._doxyxml_dir, path)) as f:
                doxyxml = ElementTree.parse(f)
                self._ns2doxyxml[namespace] = doxyxml

        nodes = doxyxml.findall(
            f"compounddef/sectiondef/memberdef/name[.='{name}']/..")
        if len(nodes) == 0:
            nodes = self._file_doxyxml.findall(
                f"compounddef/sectiondef/memberdef/name[.='{name}']/..")
        candidates = []
        for node in nodes:
            # Process a function or a typedef.
            params = None
            d = Definition(name, node=node)
            if d.kind == 'function':
                params = convert_params(node)
                node_param_str = ', '.join([p.type for p in params])
                if param_str and param_str != node_param_str:
                    candidates.append(f'{name}({node_param_str})')
                    continue
            elif d.kind == 'define':
                params = []
                for p in node.findall('param'):
                    param = Definition(p.find('defname').text, kind='param')
                    param.type = None
                    params.append(param)
            d.type = convert_type(node.find('type'))
            d.template_params = convert_template_params(node)
            d.params = params
            convert_return_type(d, node)
            d.desc = get_description(node)
            return d

        cls = doxyxml.findall(f"compounddef/innerclass[.='{qual_name}']")
        if not cls:
            raise Exception(f'Cannot find {identifier}. Candidates: {candidates}')
        return self.collect_compound(identifier, cls)

    def render(self, d: Definition, config: dict) -> str:
        if d.id is not None:
            self.do_heading('', 0, id=d.id)
        text = '<div class="docblock">\n'
        text += render_decl(d)
        text += '<div class="docblock-desc">\n'
        text += doxyxml2html(d.desc)
        if d.members is not None:
            for m in d.members:
                text += self.render(m, config)
        text += '</div>\n'
        text += '</div>\n'
        return text


def get_handler(theme: str, custom_templates: Optional[str] = None,
                **_config: Any) -> CxxHandler:
    """Return an instance of `CxxHandler`.

    Arguments:
        theme: The theme to use when rendering contents.
        custom_templates: Directory containing custom templates.
        **_config: Configuration passed to the handler.
    """
    return CxxHandler(theme=theme, custom_templates=custom_templates)



#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2012-2023 KiCad Developers
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, you may find one here:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# or you may search the http://www.gnu.org website for the version 2 license,
# or you may write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#

from pcbnew import *
import FootprintWizardBase
import pcbnew

class TouchSliderWizard(FootprintWizardBase.FootprintWizard):

    def GetName(self):
        """
        Return footprint name.
        This is specific to each footprint class, you need to implement this
        """
        return 'Touch Slider'

    def GetDescription(self):
        """
        Return footprint description.
        This is specific to each footprint class, you need to implement this
        """
        return 'Capacitive Touch Slider wizard'

    def GetValue(self):
        return "TouchSlider-{s}_{x:g}x{y:g}mm".format(
            s = self.pads['steps'],
            x = pcbnew.ToMM(self.pads['length']),
            y = pcbnew.ToMM(self.pads['width'])
            )

    def GenerateParameterList(self):
        self.AddParam("Pads", "steps", self.uInteger, 4, min_value=2)
        self.AddParam("Pads", "bands", self.uInteger, 2, min_value=1)
        self.AddParam("Pads", "width", self.uMM, 10)
        self.AddParam("Pads", "length", self.uMM, 50)
        self.AddParam("Pads", "clearance", self.uMM, 1)

    @property
    def pads(self):
        return self.parameters['Pads']

    # build a rectangular pad
    def smdRectPad(self,module,size,pos,name):
        pad = PAD(module)
        pad.SetSize(size)
        pad.SetShape(PAD_SHAPE_RECT)
        pad.SetAttribute(PAD_ATTRIB_SMD)
        pad.SetLayerSet(pad.ConnSMDMask())
        pad.SetPosition(pos)
        pad.SetName(name)
        return pad


    def smdTrianglePad(self,module,size,pos,name,up_down=1,left_right=0):
        pad = PAD(module)
        pad.SetSize(VECTOR2I(size[0],size[1]))
        pad.SetShape(PAD_SHAPE_TRAPEZOID)
        pad.SetAttribute(PAD_ATTRIB_SMD)
        pad.SetLayerSet(pad.ConnSMDMask())
        pad.SetPosition(pos)
        pad.SetName(name)
        pad.SetDelta(VECTOR2I(left_right*size[1],up_down*size[0]))
        return pad


    # This method checks the parameters provided to wizard and set errors
    def CheckParameters(self):
        #TODO - implement custom checks
        pass

    # The start pad is made of a rectangular pad plus a couple of
    # triangular pads facing tips on the middle/right of the first
    # rectangular pad
    def AddStartPad(self,position,touch_width,step_length,clearance,name):
        module = self.module
        step_length = step_length - clearance
        size_pad = VECTOR2I( (int)(step_length/2.0+(step_length/3)),(int)(touch_width) )
        pad = self.smdRectPad(module,size_pad,(VECTOR2I(position)-VECTOR2I( (int)(step_length/6),0)),name)
        module.Add(pad)

        size_pad = VECTOR2I((int)(step_length/2.0),(int)(touch_width))

        tp = self.smdTrianglePad(module,VECTOR2I(size_pad[0], (int)(size_pad[1]/2)),
                                 VECTOR2I(position)+VECTOR2I( (int)(size_pad[0]/2), (int)(size_pad[1]/4) ),
                                name)
        module.Add(tp)
        tp = self.smdTrianglePad(module,VECTOR2I(size_pad[0],(int)(size_pad[1]/2) ),
                                 position+VECTOR2I( (int)(size_pad[0]/2), (int)(-size_pad[1]/4) ),
                                name
                                ,-1)
        module.Add(tp)

    # compound a "start pad" shape plus a triangle on the left, pointing to
    # the previous touch-pad
    def AddMiddlePad(self,position,touch_width,step_length,clearance,name):
        module = self.module
        step_length = step_length - clearance
        size_pad = VECTOR2I((int)(step_length/2.0),(int)(touch_width))

        size_pad = VECTOR2I((int)(step_length/2.0),(int)(touch_width))
        pad = self.smdRectPad(module,size_pad,position,name)
        module.Add(pad)

        tp = self.smdTrianglePad(module,VECTOR2I(size_pad[0],(int)(size_pad[1]/2)),
                                 position+VECTOR2I( (int)(size_pad[0]/2), (int)(size_pad[1]/4) ),
                                name)
        module.Add(tp)
        tp = self.smdTrianglePad(module,VECTOR2I(size_pad[0],(int)(size_pad[1]/2)),
                                 position+VECTOR2I( (int)(size_pad[0]/2), (int)(-size_pad[1]/4) ),
                                name
                                ,-1)
        module.Add(tp)

        tp = self.smdTrianglePad(module,VECTOR2I(size_pad[0],(int)(size_pad[1]/2)),
                                        position+VECTOR2I( -size_pad[0],0),
                                        name,
                                        0,
                                        -1)
        module.Add(tp)


    def AddFinalPad(self,position,touch_width,step_length,clearance,name):
        module = self.module
        step_length = step_length - clearance
        size_pad = VECTOR2I((int)(step_length/2.0),(int)(touch_width))

        pad = self.smdRectPad(module,
                              VECTOR2I( (int)(size_pad[0]+(step_length/3)),size_pad[1]),
                              position+VECTOR2I( (int)(step_length/6),0),
                              name)
        module.Add(pad)

        tp = self.smdTrianglePad(module,VECTOR2I(size_pad[0],(int)(size_pad[1]/2)),
                                        position+VECTOR2I(-size_pad[0],0),
                                        name,
                                        0,
                                        -1)
        module.Add(tp)

    def AddStrip(self,pos,steps,touch_width,step_length,touch_clearance):
        self.AddStartPad(pos,touch_width,step_length,touch_clearance,"1")

        for n in range(2,steps):
            pos = pos + VECTOR2I((int)(step_length),0)
            self.AddMiddlePad(pos,touch_width,step_length,touch_clearance,str(n))

        pos = pos + VECTOR2I((int)(step_length),0)
        self.AddFinalPad(pos,touch_width,step_length,touch_clearance,str(steps))

    # build the footprint from parameters
    # FIX ME: the X and Y position of the footprint can be better.
    def BuildThisFootprint(self):

        steps             = self.pads["steps"]
        bands             = self.pads["bands"]
        touch_width       = self.pads["width"]
        touch_length      = self.pads["length"]
        touch_clearance   = self.pads["clearance"]

        step_length = float(touch_length) / float(steps)

        t_size = self.GetTextSize()
        w_text = self.draw.GetLineThickness()
        ypos = touch_width/2 + t_size/2 + w_text
        self.draw.Value(0, -ypos, t_size)
        ypos += t_size + w_text*2
        self.draw.Reference(0, -ypos, t_size)

        # set SMD attribute
        self.module.SetAttributes(FP_SMD)

        # starting pad
        band_width = touch_width/bands

        xpos = -0.5 * (steps - 1) * step_length
        ypos = -0.5 * (bands - 1) * band_width

        pos = VECTOR2I( int(xpos), int(ypos) )

        for b in range(bands):
            self.AddStrip(pos,steps,band_width,step_length,touch_clearance)
            pos += VECTOR2I(0,(int)(band_width))

TouchSliderWizard().register()




from __future__ import annotations

import gc

import pytest

from pybind11_tests import class_release_gil_before_calling_cpp_dtor as m


@pytest.mark.parametrize(
    ("probe_type", "unique_key", "expected_result"),
    [
        (m.ProbeType0, "without_manipulating_gil", "1"),
        (m.ProbeType1, "release_gil_before_calling_cpp_dtor", "0"),
    ],
)
def test_gil_state_check_results(probe_type, unique_key, expected_result):
    probe_type(unique_key)
    gc.collect()
    result = m.PopPyGILState_Check_Result(unique_key)
    assert result == expected_result



#!/usr/bin/env python

# Regenerate the icons in resources/linux/mime based on the mime files
# and the original icons in bitmaps_png. Must be run from the scripts
# folder for the relative paths to work.
#
# This script assumes Inkscape is installed and in the PATH

import os, glob
import xml.etree.ElementTree as ET
from shutil import copyfile, rmtree
from subprocess import call

ICON_SOURCES = "../../resources/bitmaps_png/sources/"
DEST_FOLDER = "../../resources/linux/mime"

def icon_sourcename(icon):
    return ICON_SOURCES+"/icon_%s.svg" % icon


# Get a list of the applications we will install, their icons and mimes
app_icons = {}
for desktopfile in glob.glob(DEST_FOLDER+"/applications/*.desktop"):
    icon = None
    mimes = []
    for line in open(desktopfile):
        keypair = map(str.strip, line.split("="))
        if len(keypair) != 2: continue
        key, value = keypair
        if key == "Icon":
            icon = value
        elif key ==  "MimeType":
            mimes = [x.strip() for x in value.split(";") if str.strip(x)]
    if icon is None:
        print "WARNING: file '", desktopfile, "' contains no Icon entry, corrupted?"
        continue
    else:
        app_icons[icon] = mimes;

# Obtain the mime types we provide from the mime package XML
MIME_PACKAGE = DEST_FOLDER+"/mime/packages/kicad-kicad.xml"
mimepkg_root = ET.parse(MIME_PACKAGE).getroot()
mimepkg_mimetypes = [n.attrib['type'] for n in mimepkg_root]

# Reconcile mime types
mime_icons = {}
for mime in mimepkg_mimetypes:
    for icon, mimes in app_icons.iteritems():
        if mime in mimes:
            mime_icons[mime.replace('/','-')] = icon
            break
    else:
        print "WARNING: mimetype'", mime,"' is provided in the package, but no app is associated with it."


RESOLUTIONS = [16,22,24,32,48,64,128]


rmtree(DEST_FOLDER+'/icons')
os.makedirs(DEST_FOLDER+'/icons/hicolor/scalable/apps')
os.makedirs(DEST_FOLDER+'/icons/hicolor/scalable/mimetypes')
for r in RESOLUTIONS:
        os.makedirs(DEST_FOLDER+'/icons/hicolor/%ix%i/apps' % (r,r))
        os.makedirs(DEST_FOLDER+'/icons/hicolor/%ix%i/mimetypes' % (r,r))

for icon in app_icons.keys():
    copyfile(icon_sourcename(icon),
             DEST_FOLDER+"/icons/hicolor/scalable/apps/%s.svg" % icon)
    for r in RESOLUTIONS:
        call(['inkscape', '-f', icon_sourcename(icon),
              '-e', DEST_FOLDER+'/icons/hicolor/%ix%i/apps/%s.png' % (r, r, icon),
              '-w', str(r), '-h', str(r), '--export-area-snap'])


for mime, icon in mime_icons.iteritems():
    copyfile(icon_sourcename(icon),
             DEST_FOLDER+"/icons/hicolor/scalable/mimetypes/%s.svg" % mime)
    for r in RESOLUTIONS:
        call(['inkscape', '-f', icon_sourcename(icon),
              '-e', DEST_FOLDER+'/icons/hicolor/%ix%i/mimetypes/%s.png' % (r, r, mime),
              '-w', str(r), '-h', str(r), '--export-area-snap'])



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_unique_ptr_member as m


def test_make_unique_pointee():
    obj = m.make_unique_pointee()
    assert obj.get_int() == 213


@pytest.mark.parametrize(
    "give_up_ownership_via",
    ["give_up_ownership_via_unique_ptr", "give_up_ownership_via_shared_ptr"],
)
def test_pointee_and_ptr_owner(give_up_ownership_via):
    obj = m.pointee()
    assert obj.get_int() == 213
    owner = m.ptr_owner(obj)
    with pytest.raises(ValueError, match="Python instance was disowned"):
        obj.get_int()
    assert owner.is_owner()
    reclaimed = getattr(owner, give_up_ownership_via)()
    assert not owner.is_owner()
    assert reclaimed.get_int() == 213



from __future__ import annotations

import pytest

from pybind11_tests import vector_unique_ptr_member as m


@pytest.mark.parametrize("num_elems", range(3))
def test_create(num_elems):
    vo = m.VectorOwner.Create(num_elems)
    assert vo.data_size() == num_elems


def test_cast():
    vo = m.VectorOwner.Create(0)
    assert m.py_cast_VectorOwner_ptr(vo) is vo



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

#  last change: 2021, Jul 19.

import pcbnew
import FootprintWizardBase

# Additional import for QRCode
# see https://github.com/kazuhikoarase/qrcode-generator/blob/master/python/qrcode.py
import kicad_qrcode as qrcode  # TODO: local qrcode package is preferred, so we renamed it

# Max type number for the QRCode is 10, this is restricted by the upstream library
# https://github.com/kazuhikoarase/qrcode-generator
_PYTHON_QR_LIB_MAX_TYPE_NO = len(qrcode.QRUtil.MAX_LENGTH)

class QRCodeWizard(FootprintWizardBase.FootprintWizard):
    GetName = lambda self: '2D Barcode QRCode'
    GetDescription = lambda self: 'QR Code barcode generator'
    GetReferencePrefix = lambda self: 'QR***'
    GetValue = lambda self: self.module.Value().GetText()

    def GenerateParameterList(self):
        # TODO: add params for ECLevel and qrcode mode (num/alnum/bytes)
        self.AddParam("Barcode", "Qr Pixel Width", self.uMM, 0.5)
        self.AddParam(
            "Barcode",
            "Border Margin (Px)",
            self.uInteger,
            0,
            min_value=0,
        )
        # ErrorCorrectLevel: L = 7%, M = 15% Q = 25% H = 30%
        self.AddParam(
            "Barcode",
            "Error Correction Level",
            self.uString,
            "M",
            hint="One of L(7%), M(15%), Q(25%), H(30%)"
        )
        self.AddParam(
            "Barcode",
            "Type Number",
            self.uInteger,
            0,
            min_value=0,
            max_value=_PYTHON_QR_LIB_MAX_TYPE_NO,
            hint="Set to 0 for autodetect (will use the smallest possible)",
        )
        self.AddParam("Barcode", "Contents", self.uString, 'Example')
        self.AddParam("Barcode", "Negative", self.uBool, False)
        self.AddParam("Barcode", "Use SilkS layer", self.uBool, False)
        self.AddParam("Barcode", "Use Cu layer", self.uBool, True)
        self.AddParam("Caption", "Enabled", self.uBool, True)
        self.AddParam("Caption", "Height", self.uMM, 1.0)
        self.AddParam("Caption", "Width", self.uMM, 1.0)
        self.AddParam("Caption", "Thickness", self.uMM, 0.15)


    def CheckParameters(self):
        self.Barcode = self.parameters['Barcode']['Contents']
        self.RawECLevel = str(self.parameters['Barcode']['Error Correction Level']).upper()
        self.TypeNumber = self.parameters['Barcode']['Type Number']
        self.X = self.parameters['Barcode']['Qr Pixel Width']
        self.negative = self.parameters['Barcode']['Negative']
        self.UseSilkS = self.parameters['Barcode']['Use SilkS layer']
        self.UseCu = self.parameters['Barcode']['Use Cu layer']
        self.border = int(self.parameters['Barcode']['Border Margin (Px)'])
        self.textHeight = int(self.parameters['Caption']['Height'])
        self.textThickness = int(self.parameters['Caption']['Thickness'])
        self.textWidth = int(self.parameters['Caption']['Width'])
        self.module.Value().SetText(str(self.Barcode))


        if not (len(self.RawECLevel) == 1 and self.RawECLevel in "LMQH"):
            self.GetParam("Barcode", "Error Correction Level").AddError(
                '"Error Correction Level" must be one of L(7%), M(15%), Q(25%), H(30%)'
            )
            self.ECLevel = qrcode.ErrorCorrectLevel.M
        else:
            self.ECLevel = getattr(qrcode.ErrorCorrectLevel, self.RawECLevel)


        # Check if the content is too long
        # technically we don't need this conversion (TypeNumber=0 will be
        # regarded as max) but this isn't part of the documented API, so we'll
        # convert to be safe
        max_type_num = (
            _PYTHON_QR_LIB_MAX_TYPE_NO
            if self.TypeNumber == 0
            else self.TypeNumber
        )
        max_length = qrcode.QRUtil.getMaxLength(
            max_type_num,
            qrcode.Mode.MODE_8BIT_BYTE,
            self.ECLevel
        )
        if len(self.Barcode) > max_length:
            self.GetParam("Barcode", "Contents").AddError(
                f"Content too long ({len(self.Barcode)} > {max_length}) for"
                " the provided parameters."
            )


    def drawPixelSquareArea( self, layer, size, xposition, yposition):
        # creates a PCB_SHAPE of rectangle type. The rectangle is square
        rectangle = pcbnew.PCB_SHAPE(self.module)
        rectangle.SetShape(pcbnew.S_RECT)
        rectangle.SetWidth( 0 )
        rectangle.SetFilled( True )
        rectangle.SetLayer(layer)
        halfsize = int(size/2)
        rectangle.SetStartX( -halfsize+xposition )
        rectangle.SetStartY( -halfsize+yposition )
        rectangle.SetEndX( halfsize+xposition )
        rectangle.SetEndY( halfsize+yposition )
        return rectangle


    def _drawQrPixel(self, xposition, yposition):
        # build a rectangular pad as a dot on copper layer,
        # and a rectangle (a square) on silkscreen
        if self.UseCu:
            pad = pcbnew.PAD(self.module)
            pad.SetSize(pcbnew.VECTOR2I(self.X, self.X))
            pad_pos = pcbnew.VECTOR2I(xposition,yposition)
            pad.SetPosition(pad_pos)
            pad.SetShape(pcbnew.PAD_SHAPE_RECT)
            pad.SetAttribute(pcbnew.PAD_ATTRIB_SMD)
            pad.SetName("")
            layerset = pcbnew.LSET()
            layerset.AddLayer(pcbnew.F_Cu)
            pad.SetLayerSet( layerset )
            self.module.Add(pad)
        if self.UseSilkS:
            rectangle=self.drawPixelSquareArea(pcbnew.F_SilkS, self.X, xposition, yposition)
            self.module.Add(rectangle)


    def BuildThisFootprint(self):
        # Build QRCode
        # Auto detect type number
        if self.TypeNumber == 0:
            self.qr = qrcode.QRCode.getMinimumQRCode(self.Barcode, self.ECLevel)
        # Manually specified type number
        else:
            self.qr = qrcode.QRCode()
            self.qr.setTypeNumber(self.TypeNumber)
            self.qr.setErrorCorrectLevel(self.ECLevel)
            self.qr.addData(self.Barcode)
            self.qr.make()

        # render QRCode
        if self.border >= 0:
            # Adding border: Create a new array larger than the self.qr.modules
            sz = self.qr.modules.__len__() + (self.border * 2)
            arrayToDraw = [ [ 0 for a in range(sz) ] for b in range(sz) ]
            lineposition = self.border

            for i in self.qr.modules:
                columnposition = self.border
                for j in i:
                    arrayToDraw[lineposition][columnposition] = j
                    columnposition += 1
                lineposition += 1
        else:
            # No border: using array as is
            arrayToDraw = self.qr.modules

        # used many times...
        half_number_of_elements = arrayToDraw.__len__() / 2
        area_width =  arrayToDraw.__len__()*self.X + self.border*2

        # Center position of QrCode
        yposition = - int(half_number_of_elements * self.X) + int(self.X/2)
        area_height =  arrayToDraw.__len__()*self.X + self.border*2

        for line in arrayToDraw:
            xposition = - int(half_number_of_elements * self.X) + int(self.X/2)
            for pixel in line:
                # Trust table for drawing a pixel
                # Negative is a boolean;
                # each pixel is a boolean (need to draw of not)
                # Negative | Pixel | Result
                #        0 |     0 | 0
                #        0 |     1 | 1
                #        1 |     0 | 1
                #        1 |     1 | 0
                # => Draw as Xor
                if self.negative != pixel: # Xor...
                    self._drawQrPixel(xposition, yposition)
                xposition += self.X
            yposition += self.X

        # Add value field
        textPosition = int((self.textHeight) + ((1 + half_number_of_elements) * self.X))
        pos = pcbnew.VECTOR2I(0, - textPosition)
        self.module.Value().SetPosition(pos)
        self.module.Value().SetTextHeight(self.textHeight)
        self.module.Value().SetTextWidth(self.textWidth)
        self.module.Value().SetTextThickness(self.textThickness)

        # Add Reference field
        pos = pcbnew.VECTOR2I(0, textPosition)
        self.module.Reference().SetPosition(pos)
        self.module.Reference().SetTextHeight(self.textHeight)
        self.module.Reference().SetTextWidth(self.textWidth)
        self.module.Reference().SetTextThickness(self.textThickness)
        self.module.Value().SetLayer(pcbnew.F_SilkS)

        #build the footprint courtyard
        self.draw.SetLayer( pcbnew.F_CrtYd )
        self.draw.SetLineThickness( pcbnew.FromMM( 0.05 ) ) #Default per KLC F5.1 as of 12/2018
        cr_margin = pcbnew.FromMM( 0.1 )
        self.draw.Box(0, 0, area_width + cr_margin*2, area_height + cr_margin*2)

        #build the footprint solder mask: the solder mask covers all copper pads
        if self.UseCu:
            self.draw.SetLineThickness( 0 )
            sm_margin = pcbnew.FromMM( 0.05 )
            rectangle=self.drawPixelSquareArea(pcbnew.F_Mask, area_width + sm_margin*2, 0, 0)
            self.module.Add(rectangle)


QRCodeWizard().register()
